"""``runway new`` command."""
# docs: file://./../../../docs/source/commands.rst
import locale
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import click

from .. import options

if TYPE_CHECKING:
    from ..._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__.replace("._", ".")))
RUNWAY_YML = """---
# See full syntax at https://docs.onica.com/projects/runway
deployments:
  - modules:
      - path: sampleapp.cfn
      - path: sampleapp.tf
      # - etc...
    regions:
      - us-east-1
"""


@click.command("new", short_help="create runway.yml")
@options.debug
@options.no_color
@options.verbose
@click.pass_context
def new(ctx: click.Context, **_: Any) -> None:
    """Create an example runway.yml file in the correct directory."""
    runway_yml = Path.cwd() / "runway.yml"

    LOGGER.verbose("checking for preexisting runway.yml file...")
    if runway_yml.is_file():
        LOGGER.error(
            "There is already a %s file in the current directory", runway_yml.name
        )
        ctx.exit(1)

    runway_yml.write_text(
        RUNWAY_YML, encoding=locale.getpreferredencoding(do_setlocale=False)
    )
    LOGGER.success("runway.yml generated")
    LOGGER.notice(
        "See addition getting started information at "
        "https://docs.onica.com/projects/runway/page/getting_started.html"
    )



"""Format path.

.. Derived from software distributed by Amazon.com, Inc - http://aws.amazon.com/apache2.0/
   https://github.com/aws/aws-cli/blob/83b43782dd/awscli/customizations/s3/fileformat.py

"""
from __future__ import annotations

import os
from pathlib import Path
from typing import Tuple

from typing_extensions import Literal, TypedDict

SupportedPathType = Literal["local", "s3"]
FormattedPathDetails = TypedDict(
    "FormattedPathDetails", path=str, type=SupportedPathType
)
FormatPathResult = TypedDict(
    "FormattedPaths",
    dest=FormattedPathDetails,
    dir_op=bool,
    src=FormattedPathDetails,
    use_src_name=bool,
)


class FormatPath:
    """Path format base class."""

    @classmethod
    def format(cls, src: str, dest: str) -> FormatPathResult:
        """Format the source and destination for use in the file factory."""
        src_type, src_path = cls.identify_path_type(src)
        dest_type, dest_path = cls.identify_path_type(dest)
        format_table = {"s3": cls.format_s3_path, "local": cls.format_local_path}

        src_path = format_table[src_type](src_path)[0]
        dest_path, use_src_name = format_table[dest_type](dest_path)

        return {
            "dest": {"path": dest_path, "type": dest_type},
            "dir_op": True,
            "src": {"path": src_path, "type": src_type},
            "use_src_name": use_src_name,
        }

    @staticmethod
    def format_local_path(path: str, dir_op: bool = True) -> Tuple[str, bool]:
        """Format the path of local files.

        Returns whether the destination will keep its own name or take the
        source's name along with the edited path.

        Formatting Rules:
            1. If a destination file is taking on a source name, it must end
               with the appropriate operating system separator

        General Options:
            1. If the operation is on a directory, the destination file will
               always use the name of the corresponding source file.
            2. If the path of the destination exists and is a directory it
               will always use the name of the source file.
            3. If the destination path ends with the appropriate operating
               system separator but is not an existing directory, the
               appropriate directories will be made and the file will use the
               source's name.
            4. If the destination path does not end with the appropriate
               operating system separator and is not an existing directory, the
               appropriate directories will be created and the file name will
               be of the one provided.

        """
        full_path = Path(path).resolve()
        if full_path.is_dir() or dir_op:
            return f"{full_path}{os.sep}", True
        if path.endswith(os.sep):
            return f"{full_path}{os.sep}", True
        return str(full_path), False

    @staticmethod
    def format_s3_path(path: str, dir_op: bool = True) -> Tuple[str, bool]:
        """Format the path of S3 files.

        Returns whether the destination will keep its own name or take the
        source's name along with the edited path.

        Formatting Rules:
            1. If a destination file is taking on a source name, it must end
               with a forward slash.

        General Options:
            1. If the operation is on objects under a common prefix,
               the destination file will always use the name of the
               corresponding source file.
            2. If the path ends with a forward slash, the appropriate prefixes
               will be formed and will use the name of the source.
            3. If the path does not end with a forward slash, the appropriate
               prefix will be formed but use the the name provided as opposed
               to the source name.

        """
        if dir_op:
            if not path.endswith("/"):
                path += "/"
            return path, True
        if path.endswith("/"):
            return path, True
        return path, False

    @staticmethod
    def identify_path_type(path: str) -> Tuple[SupportedPathType, str]:
        """Parse path.

        Args:
            path: Path to identify.

        Returns:
            Path type & raw path that has been resolved and/or stripped of prefix.

        """
        if path.startswith("s3://"):
            return "s3", path[5:]
        return "local", path



"""Context type definitions."""
from __future__ import annotations

from typing_extensions import TypedDict


class PersistentGraphLocation(TypedDict, total=False):
    """CFNgin persistent graph location."""

    Bucket: str
    Key: str



"""Runway deploy environment object."""
from __future__ import annotations

import json
import logging
import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional, cast

import click

from ...compat import cached_property
from ...mixins import DelCachedPropMixin
from ...type_defs import EnvVarsAwsCredentialsTypeDef
from ...utils import AWS_ENV_VARS

try:  # will raise an import error if git is not in the current path
    import git
    from git.exc import InvalidGitRepositoryError
except ImportError:  # cov: ignore
    git = object  # pylint: disable=invalid-name
    InvalidGitRepositoryError = AttributeError

if TYPE_CHECKING:
    from ..._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__.replace("._", ".")))


class DeployEnvironment(DelCachedPropMixin):
    """Runway deploy environment."""

    __name: Optional[str]
    _ignore_git_branch: bool

    name_derived_from: Optional[str]
    root_dir: Path

    def __init__(
        self,
        *,
        environ: Optional[Dict[str, str]] = None,
        explicit_name: Optional[str] = None,
        ignore_git_branch: bool = False,
        root_dir: Optional[Path] = None,
    ) -> None:
        """Instantiate class.

        Args:
            environ: Environment variables.
            explicit_name: Explicitly provide the deploy environment name.
            ignore_git_branch: Ignore the git branch when determining the deploy
                environment name.
            root_dir: Root directory of the project.

        """
        self.__name = explicit_name
        self._ignore_git_branch = ignore_git_branch
        self.name_derived_from = "explicit" if self.__name else None
        self.root_dir = root_dir or Path.cwd()
        self.vars = environ or os.environ.copy()

    @property
    def aws_credentials(self) -> EnvVarsAwsCredentialsTypeDef:
        """Get AWS credentials from environment variables."""
        return EnvVarsAwsCredentialsTypeDef(
            **{name: self.vars[name] for name in AWS_ENV_VARS if self.vars.get(name)}
        )

    @property
    def aws_profile(self) -> Optional[str]:
        """Get AWS profile from environment variables."""
        return self.vars.get("AWS_PROFILE")

    @aws_profile.setter
    def aws_profile(self, profile_name: str) -> None:
        """Set AWS profile in the environment."""
        self._update_vars({"AWS_PROFILE": profile_name})

    @property
    def aws_region(self) -> str:
        """Get AWS region from environment variables."""
        return self.vars.get(
            "AWS_REGION", self.vars.get("AWS_DEFAULT_REGION", "us-east-1")
        )

    @aws_region.setter
    def aws_region(self, region: str) -> None:
        """Set AWS region environment variables."""
        self._update_vars({"AWS_DEFAULT_REGION": region, "AWS_REGION": region})

    @cached_property
    def branch_name(self) -> Optional[str]:
        """Git branch name."""
        if isinstance(git, type):
            LOGGER.debug(
                "failed to import git; ensure git is your path and "
                "executable to read the branch name"
            )
            return None
        try:
            LOGGER.debug("getting git branch name...")
            return git.Repo(  # type: ignore
                str(self.root_dir), search_parent_directories=True
            ).active_branch.name
        except InvalidGitRepositoryError:
            return None
        except TypeError:
            LOGGER.warning("Unable to retrieve the current git branch name!")
            LOGGER.warning(
                "Typically this occurs when operating in a "
                "detached-head state (e.g. what Jenkins uses when "
                "checking out a git branch). Set the "
                "DEPLOY_ENVIRONMENT environment variable to the "
                'name of the logical environment (e.g. "export '
                'DEPLOY_ENVIRONMENT=dev") to bypass this error.'
            )
            sys.exit(1)

    @property
    def ci(self) -> bool:
        """Return CI status.

        Returns:
            bool

        """
        return "CI" in self.vars

    @ci.setter
    def ci(self, value: Any) -> None:
        """Set the value of CI."""
        if value:
            self._update_vars({"CI": "1"})
        else:
            self.vars.pop("CI", None)

    @ci.deleter
    def ci(self) -> None:
        """Delete the value of CI."""
        self.vars.pop("CI", None)

    @property
    def debug(self) -> bool:
        """Get debug setting from the environment."""
        return "DEBUG" in self.vars

    @debug.setter
    def debug(self, value: Any) -> None:
        """Set the value of DEBUG."""
        if value:
            self._update_vars({"DEBUG": "1"})
        else:
            self.vars.pop("DEBUG", None)

    @property
    def ignore_git_branch(self) -> bool:
        """Whether to ignore git branch when determining name."""
        return self._ignore_git_branch

    @ignore_git_branch.setter
    def ignore_git_branch(self, value: bool) -> None:
        """Set the value of ignore_git_branch.

        Cached name is deleted when changing this value.

        """
        if self._ignore_git_branch != value:
            self._ignore_git_branch = value
            self._del_cached_property("name")
            LOGGER.debug(
                "value of ignore_git_branch has changed; "
                "cleared cached name so it can be determined again"
            )

    @property
    def max_concurrent_cfngin_stacks(self) -> int:
        """Max number of CFNgin stacks that can be deployed concurrently.

        This property can be set by exporting
        ``RUNWAY_MAX_CONCURRENT_CFNGIN_STACKS``. If no value is specified, the
        value will be constrained based on the underlying graph.

        Returns:
            Value from environment variable or ``0``.

        """
        return int(self.vars.get("RUNWAY_MAX_CONCURRENT_CFNGIN_STACKS", "0"))

    @max_concurrent_cfngin_stacks.setter
    def max_concurrent_cfngin_stacks(self, value: int) -> None:
        """Set RUNWAY_MAX_CONCURRENT_CFNGIN_STACKS."""
        self._update_vars({"RUNWAY_MAX_CONCURRENT_CFNGIN_STACKS": str(value)})

    @property
    def max_concurrent_modules(self) -> int:
        """Max number of modules that can be deployed to concurrently.

        This property can be set by exporting ``RUNWAY_MAX_CONCURRENT_MODULES``.
        If no value is specified, ``min(61, os.cpu_count())`` is used.

        On Windows, this must be equal to or lower than ``61``.

        **IMPORTANT:** When using ``parallel_regions`` and ``child_modules``
        together, please consider the nature of their relationship when
        manually setting this value. (``parallel_regions * child_modules``)

        Returns:
            Value from environment variable or ``min(61, os.cpu_count())``

        """
        value = self.vars.get("RUNWAY_MAX_CONCURRENT_MODULES")

        if value:
            return int(value)
        return min(61, os.cpu_count() or 61)

    @max_concurrent_modules.setter
    def max_concurrent_modules(self, value: int) -> None:
        """Set RUNWAY_MAX_CONCURRENT_MODULES."""
        self._update_vars({"RUNWAY_MAX_CONCURRENT_MODULES": str(value)})

    @property
    def max_concurrent_regions(self) -> int:
        """Max number of regions that can be deployed to concurrently.

        This property can be set by exporting ``RUNWAY_MAX_CONCURRENT_REGIONS``.
        If no value is specified, ``min(61, os.cpu_count())`` is used.

        On Windows, this must be equal to or lower than ``61``.

        **IMPORTANT:** When using ``parallel_regions`` and ``child_modules``
        together, please consider the nature of their relationship when
        manually setting this value. (``parallel_regions * child_modules``)

        Returns:
            Value from environment variable or ``min(61, os.cpu_count())``

        """
        value = self.vars.get("RUNWAY_MAX_CONCURRENT_REGIONS")

        if value:
            return int(value)
        return min(61, os.cpu_count() or 61)

    @max_concurrent_regions.setter
    def max_concurrent_regions(self, value: int) -> None:
        """Set RUNWAY_MAX_CONCURRENT_REGIONS."""
        self._update_vars({"RUNWAY_MAX_CONCURRENT_REGIONS": str(value)})

    @cached_property
    def name(self) -> str:
        """Deploy environment name."""
        if self.__name:
            name = self.__name
        elif not self.ignore_git_branch and self.branch_name:
            self.name_derived_from = self.name_derived_from or "branch"
            name = self._parse_branch_name()
        else:
            self.name_derived_from = "directory"
            if self.root_dir.name.startswith("ENV-"):
                LOGGER.verbose(
                    'stripped "ENV-" from the directory name "%s"', self.root_dir.name
                )
                name = self.root_dir.name[4:]
            else:
                name = self.root_dir.name
        if not name:
            raise ValueError("could not determine deploy environment name")
        if self.vars.get("DEPLOY_ENVIRONMENT") != name:
            self._update_vars({"DEPLOY_ENVIRONMENT": name})
        return name

    @property
    def verbose(self) -> bool:
        """Get verbose setting from the environment."""
        return "VERBOSE" in self.vars

    @verbose.setter
    def verbose(self, value: Any) -> None:
        """Set the value of VERBOSE."""
        if value:
            self._update_vars({"VERBOSE": "1"})
        else:
            self.vars.pop("VERBOSE", None)

    def copy(self) -> DeployEnvironment:
        """Copy the contents of this object into a new instance.

        Returns:
            DeployEnvironment: New instance with the same contents.

        """
        LOGGER.debug("creating a copy of the deploy environment...")
        obj = self.__class__(
            environ=self.vars.copy(),
            explicit_name=self.name,
            ignore_git_branch=self._ignore_git_branch,
            root_dir=self.root_dir,
        )
        obj.name_derived_from = self.name_derived_from
        return obj

    def log_name(self) -> None:
        """Output name to log."""
        name = self.name  # resolve if not already resolved
        if self.name_derived_from == "explicit":
            LOGGER.info(
                'deploy environment "%s" is explicitly defined in the environment', name
            )
            LOGGER.info(
                "if not correct, update the value or unset it to fall back "
                "to the name of the current git branch or parent directory"
            )
        elif self.name_derived_from == "branch":
            LOGGER.info(
                'deploy environment "%s" was determined from the current git branch',
                name,
            )
            LOGGER.info(
                "if not correct, update the branch name or set an override "
                "via the DEPLOY_ENVIRONMENT environment variable"
            )
        elif self.name_derived_from == "directory":
            LOGGER.info(
                'deploy environment "%s" was determined from the current directory',
                name,
            )
            LOGGER.info(
                "if not correct, update the directory name or set an "
                "override via the DEPLOY_ENVIRONMENT environment variable"
            )

    def _parse_branch_name(self) -> Optional[str]:
        """Parse branch name for use as deploy environment name."""
        if self.branch_name:
            if self.branch_name.startswith("ENV-"):
                LOGGER.verbose(
                    'stripped "ENV-" from the branch name "%s"', self.branch_name
                )
                return self.branch_name[4:]
            if self.branch_name == "master":
                LOGGER.verbose('translated branch name "master" to "common"')
                return "common"
        if not self.ci:
            LOGGER.warning('Found unexpected branch name "%s"', self.branch_name)
            result = click.prompt(
                "Deploy environment name", default=self.branch_name, type=click.STRING
            )
            if result != self.branch_name:
                self.name_derived_from = "explicit"
            return result
        return self.branch_name

    def _update_vars(self, env_vars: Dict[str, str]) -> None:
        """Update vars and log the change.

        Args:
            env_vars (Dict[str, str]): Dict to update self.vars with.

        """
        self.vars.update(env_vars)
        LOGGER.verbose("updated environment variables: %s", json.dumps(env_vars))



"""Parameters."""
from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, List, Optional, Union, cast

from pydantic import validator
from typing_extensions import Literal

from ......utils import BaseModel
from .utils import find_bucket_key

PathsType = Literal["local", "locallocal", "locals3", "s3", "s3local", "s3s3"]


class ParametersDataModel(BaseModel):
    """Parameters data model.

    Attributes:
        dest: File/object destination.
        src: File/object source.
        content_type: Explicitly provided content type.
        delete: Whether or not to delete files at the destination that are
            missing from the source location.
        dir_op: If the source location is a directory.
        dryrun: Whether this is a dry run.
        exact_timestamps: Use exact time stamp when comparing files/objects
            during sync.
        exclude: List of patterns for files/objects to exclude.
        expected_size: Expected size of transfer.
        follow_symlinks: Whether or not to follow symlinks.
        force_glacier_transfer: Force transfer even if glacier.
        guess_mime_type: Whether or not to guess content type.
        ignore_glacier_warnings: Don't show glacier warnings.
        include: List of patterns for files/objects to explicitly include.
        is_move: Whether or not the action is move.
        is_stream: Source or destination is a stream.
        no_progress: Whether to not show progress.
        only_show_errors: Whether or not to only show errors while running.
        page_size: Number of objects to list per call.
        quiet: Don't output anything.
        paths_type: Concatenated path types for source and destination.
        size_only: When comparing files/objects, only consider size.
        storage_class: S3 storage class.

    """

    dest: str
    src: str
    # these need to be set after dest & src so their validators can access the value if needed
    content_type: Optional[str] = None
    delete: bool = False
    dir_op: bool = False
    dryrun: bool = False
    exact_timestamps: bool = False
    exclude: List[str] = []
    expected_size: Optional[int] = None
    follow_symlinks: bool = False
    force_glacier_transfer: bool = False
    guess_mime_type: bool = True
    ignore_glacier_warnings: bool = False
    include: List[str] = []
    is_move: bool = False
    is_stream: bool = False
    no_progress: bool = False
    only_show_errors: bool = False
    page_size: Optional[int] = None
    paths_type: PathsType = "local"  # will be overwritten
    quiet: bool = False
    size_only: bool = False
    sse_c: Optional[str] = None
    sse_c_key: Optional[str] = None
    storage_class: Optional[str] = None

    @validator("paths_type", always=True, pre=True)
    @classmethod
    def _determine_paths_type(
        cls,
        v: Optional[str],  # pylint: disable=unused-argument
        values: Dict[str, Any],
    ) -> PathsType:
        """Determine paths type for the given src and dest."""
        # these have already been validated so it's "safe" to cast them
        dest = cast(str, values.get("dest", ""))
        src = cast(str, values.get("src", ""))
        src_type = "s3" if src.startswith("s3://") else "local"
        dest_type = "s3" if dest.startswith("s3://") else "local"
        return cast(PathsType, f"{src_type}{dest_type}")

    @validator("dest", "src", pre=True)
    @classmethod
    def _normalize_s3_trailing_slash(cls, v: str) -> str:
        """Add a trailing "/" if the root of an S3 bucket was provided."""
        if v.startswith("s3://"):
            _bucket, key = find_bucket_key(v[5:])
            if not key and not v.endswith("/"):
                v += "/"
        return v


class Parameters:
    """Initial error based on the parameters and arguments passed to sync."""

    def __init__(
        self, action: str, parameters: Union[Dict[str, Any], ParametersDataModel]
    ):
        """Instantiate class.

        Args:
            action: The name of the action.
            parameters: A dictionary of parameters.

        """
        self.action = action
        self.data = ParametersDataModel.parse_obj(parameters)
        if self.action in ["sync", "mb", "rb"]:
            self.data.dir_op = True
        if self.action == "mv":
            self.data.is_move = True
        else:
            self.data.is_move = False
        self._validate_path_args()

    def _validate_path_args(self) -> None:
        # If we're using a mv command, you can't copy the object onto itself.
        if self.action == "mv" and self._same_path():
            raise ValueError(
                f"Cannot mv a file onto itself: '{self.data.src}' - '{self.data.dest}'"
            )

        # If the operation is downloading to a directory that does not exist,
        # create the directories so no warnings are thrown during the syncing
        # process.
        if self.data.paths_type == "s3local" and self.data.dir_op:
            dest_path = Path(self.data.dest)
            if not dest_path.exists():
                dest_path.mkdir(exist_ok=True, parents=True)

    def _same_path(self) -> bool:
        """Evaluate if the src and dest are the same path."""
        if not self.data.paths_type == "s3s3":
            return False
        if self.data.src == self.data.dest:
            return True
        return False



"""Register test handlers."""
from __future__ import annotations

import logging
from typing import Dict, Type, Union, cast

from ..utils import load_object_from_string
from .handlers.base import LookupHandler
from .handlers.cfn import CfnLookup
from .handlers.ecr import EcrLookup
from .handlers.env import EnvLookup
from .handlers.random_string import RandomStringLookup
from .handlers.ssm import SsmLookup
from .handlers.var import VarLookup

RUNWAY_LOOKUP_HANDLERS: Dict[str, Type[LookupHandler]] = {}
LOGGER = logging.getLogger(__name__)


def register_lookup_handler(
    lookup_type: str, handler_or_path: Union[str, Type[LookupHandler]]
) -> None:
    """Register a lookup handler.

    Args:
        lookup_type: Name to register the handler under
        handler_or_path: a function or a path to a handler

    """
    handler = handler_or_path

    if isinstance(handler_or_path, str):
        handler = cast(type, load_object_from_string(handler_or_path))
    else:
        handler = handler_or_path

    try:
        if issubclass(handler, LookupHandler):
            RUNWAY_LOOKUP_HANDLERS[lookup_type] = handler
            return
    except Exception:  # pylint: disable=broad-except
        LOGGER.debug("failed to validate lookup handler", exc_info=True)
    raise TypeError(
        f"lookup {handler_or_path} must be a subclass of "
        "runway.lookups.handlers.base.LookupHandler"
    )


def unregister_lookup_handler(lookup_type: str) -> None:
    """Unregister the specified test type.

    This is useful when testing various lookup types if you want to unregister
    the lookup type after the test runs.

    Args:
        lookup_type: Name of the lookup type to unregister

    """
    RUNWAY_LOOKUP_HANDLERS.pop(lookup_type, None)


register_lookup_handler(CfnLookup.TYPE_NAME, CfnLookup)
register_lookup_handler(EcrLookup.TYPE_NAME, EcrLookup)
register_lookup_handler(EnvLookup.TYPE_NAME, EnvLookup)
register_lookup_handler(RandomStringLookup.TYPE_NAME, RandomStringLookup)
register_lookup_handler(SsmLookup.TYPE_NAME, SsmLookup)
register_lookup_handler(VarLookup.TYPE_NAME, VarLookup)



"""AWS Lambda Python Deployment Package."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Optional

from igittigitt import IgnoreParser

from .....compat import cached_property
from ..deployment_package import DeploymentPackage

if TYPE_CHECKING:
    from . import PythonProject


class PythonDeploymentPackage(DeploymentPackage["PythonProject"]):
    """AWS Lambda Python Deployment Package."""

    project: PythonProject

    @cached_property
    def gitignore_filter(self) -> Optional[IgnoreParser]:
        """Filter to use when zipping dependencies.

        This should be overridden by subclasses if a filter should be used.

        """
        if self.project.args.slim:
            gitignore_filter = IgnoreParser()
            gitignore_filter.add_rule(
                "**/*.dist-info*", self.project.dependency_directory
            )
            gitignore_filter.add_rule(
                "**/*.py[c|d|i|o]", self.project.dependency_directory
            )
            gitignore_filter.add_rule(
                "**/__pycache__*", self.project.dependency_directory
            )
            if self.project.args.strip:
                gitignore_filter.add_rule("**/*.so", self.project.dependency_directory)
            return gitignore_filter
        return None

    @staticmethod
    def insert_layer_dir(file_path: Path, relative_to: Path) -> Path:
        """Insert ``python`` directory into local file path for layer archive.

        Args:
            file_path: Path to local file.
            relative_to: Path to a directory that the file_path will be relative
                to in the deployment package.

        """
        return relative_to / f"python/{file_path.relative_to(relative_to)}"



"""CFNgin plan, plan components, and functions for interacting with a plan."""
from __future__ import annotations

import json
import logging
import os
import threading
import time
import uuid
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    NoReturn,
    Optional,
    OrderedDict,
    Set,
    TypeVar,
    Union,
    overload,
)

from .._logging import LogLevels, PrefixAdaptor
from ..utils import merge_dicts
from .dag import DAG, DAGValidationError, walk
from .exceptions import CancelExecution, GraphError, PersistentGraphLocked, PlanFailed
from .stack import Stack
from .status import (
    COMPLETE,
    FAILED,
    PENDING,
    SKIPPED,
    SUBMITTED,
    FailedStatus,
    SkippedStatus,
)
from .ui import ui
from .utils import stack_template_key_name

if TYPE_CHECKING:
    from ..context import CfnginContext
    from .providers.aws.default import Provider
    from .status import Status

LOGGER = logging.getLogger(__name__)

_T = TypeVar("_T")


@overload
def json_serial(obj: Set[_T]) -> List[_T]:
    ...


@overload
def json_serial(obj: Union[Dict[Any, Any], int, List[Any], str]) -> NoReturn:
    ...


def json_serial(obj: Union[Set[Any], Any]) -> Any:
    """Serialize json.

    Args:
        obj: A python object.

    Example:
        json.dumps(data, default=json_serial)

    """
    if isinstance(obj, set):
        return list(obj)
    raise TypeError


def merge_graphs(graph1: Graph, graph2: Graph) -> Graph:
    """Combine two Graphs into one, retaining steps.

    Args:
        graph1: Graph that ``graph2`` will be merged into.
        graph2: Graph that will be merged into ``graph1``.

    """
    merged_graph_dict = merge_dicts(graph1.to_dict().copy(), graph2.to_dict())
    steps = [
        graph1.steps.get(name, graph2.steps.get(name))
        for name in merged_graph_dict.keys()
    ]
    return Graph.from_steps([step for step in steps if step])


class Step:
    """State machine for executing generic actions related to stacks.

    Attributes:
        fn: Function to run to execute the step.
            This function will be ran multiple times until the step is "done".
        last_updated: Time when the step was last updated.
        logger: Logger for logging messages about the step.
        stack: the stack associated with this step
        status: The status of step.
        watch_func: Function that will be called to "tail" the step action.

    """

    fn: Optional[Callable[..., Any]]
    last_updated: float
    logger: PrefixAdaptor
    stack: Stack
    status: Status
    watch_func: Optional[Callable[..., Any]]

    def __init__(
        self,
        stack: Stack,
        *,
        fn: Optional[Callable[..., Any]] = None,
        watch_func: Optional[Callable[..., Any]] = None,
    ) -> None:
        """Instantiate class.

        Args:
            stack: The stack associated
                with this step
            fn: Function to run to execute the step.
                This function will be ran multiple times until the step is "done".
            watch_func: Function that will be called to "tail" the step action.

        """
        self.stack = stack
        self.status = PENDING
        self.last_updated = time.time()
        self.logger = PrefixAdaptor(self.stack.name, LOGGER)
        self.fn = fn
        self.watch_func = watch_func

    def run(self) -> bool:
        """Run this step until it has completed or been skipped."""
        stop_watcher = threading.Event()
        watcher = None
        if self.watch_func:
            watcher = threading.Thread(
                target=self.watch_func, args=(self.stack, stop_watcher)
            )
            watcher.start()

        try:
            while not self.done:
                self._run_once()
        finally:
            if watcher:
                stop_watcher.set()
                watcher.join()
        return self.ok

    def _run_once(self) -> Status:
        """Run a step exactly once."""
        if not self.fn:
            raise TypeError("Step.fn must be type Callable[..., Status] not None")
        try:
            status = self.fn(self.stack, status=self.status)
        except CancelExecution:
            status = SkippedStatus("canceled execution")
        except Exception as err:  # pylint: disable=broad-except
            LOGGER.exception(err)
            status = FailedStatus(reason=str(err))
        self.set_status(status)
        return status

    @property
    def name(self) -> str:
        """Name of the step.

        This is equal to the name of the stack it operates on.

        """
        return self.stack.name

    @property
    def requires(self) -> Set[str]:
        """Return a list of step names this step depends on."""
        return self.stack.requires

    @property
    def required_by(self) -> Set[str]:
        """Return a list of step names that depend on this step."""
        return self.stack.required_by

    @property
    def completed(self) -> bool:
        """Return True if the step is in a COMPLETE state."""
        return self.status == COMPLETE

    @property
    def skipped(self) -> bool:
        """Return True if the step is in a SKIPPED state."""
        return self.status == SKIPPED

    @property
    def failed(self) -> bool:
        """Return True if the step is in a FAILED state."""
        return self.status == FAILED

    @property
    def done(self) -> bool:
        """Return True if the step is finished.

        To be ``True``, status must be either COMPLETE, SKIPPED or FAILED)

        """
        return self.completed or self.skipped or self.failed

    @property
    def ok(self) -> bool:
        """Return True if the step is finished (either COMPLETE or SKIPPED)."""
        return self.completed or self.skipped

    @property
    def submitted(self) -> bool:
        """Return True if the step is SUBMITTED, COMPLETE, or SKIPPED."""
        return self.status >= SUBMITTED

    def set_status(self, status: Status) -> None:
        """Set the current step's status.

        Args:
            status: The status to set the step to.

        """
        if status is not self.status:
            LOGGER.debug("setting %s state to %s...", self.stack.name, status.name)
            self.status = status
            self.last_updated = time.time()
            if self.stack.logging:
                self.log_step()

    def complete(self) -> None:
        """Shortcut for ``set_status(COMPLETE)``."""
        self.set_status(COMPLETE)

    def log_step(self) -> None:
        """Construct a log message for a set and log it to the UI."""
        msg = self.status.name
        if self.status.reason:
            msg += f" ({self.status.reason})"
        if self.status.code == SUBMITTED.code:
            ui.log(LogLevels.NOTICE, msg, logger=self.logger)
        elif self.status.code == COMPLETE.code:
            ui.log(LogLevels.SUCCESS, msg, logger=self.logger)
        elif self.status.code == FAILED.code:
            ui.log(LogLevels.ERROR, msg, logger=self.logger)
        else:
            ui.info(msg, logger=self.logger)

    def skip(self) -> None:
        """Shortcut for ``set_status(SKIPPED)``."""
        self.set_status(SKIPPED)

    def submit(self) -> None:
        """Shortcut for ``set_status(SUBMITTED)``."""
        self.set_status(SUBMITTED)

    @classmethod
    def from_stack_name(
        cls,
        stack_name: str,
        context: CfnginContext,
        requires: Optional[Union[List[str], Set[str]]] = None,
        fn: Optional[Callable[..., Status]] = None,
        watch_func: Optional[Callable[..., Any]] = None,
    ) -> Step:
        """Create a step using only a stack name.

        Args:
            stack_name: Name of a CloudFormation stack.
            context: Context object. Required to initialize a "fake"
                :class:`runway.cfngin.stack.Stack`.
            requires: Stacks that this stack depends on.
            fn: The function to run to execute the step.
                This function will be ran multiple times until the step is "done".
            watch_func: an optional function that will be called to "tail" the
                step action.

        """
        # pylint: disable=import-outside-toplevel
        from runway.config.models.cfngin import CfnginStackDefinitionModel

        stack_def = CfnginStackDefinitionModel.construct(
            name=stack_name, requires=requires or []
        )
        stack = Stack(stack_def, context)
        return cls(stack, fn=fn, watch_func=watch_func)

    @classmethod
    def from_persistent_graph(
        cls,
        graph_dict: Union[
            Dict[str, List[str]], Dict[str, Set[str]], OrderedDict[str, Set[str]]
        ],
        context: CfnginContext,
        fn: Optional[Callable[..., Status]] = None,
        watch_func: Optional[Callable[..., Any]] = None,
    ) -> List[Step]:
        """Create a steps for a persistent graph dict.

        Args:
            graph_dict: A graph dict.
            context: Context object. Required to initialize a "fake"
                :class:`runway.cfngin.stack.Stack`.
            requires: Stacks that this stack depends on.
            fn: The function to run to execute the step.
                This function will be ran multiple times until the step is "done".
            watch_func: an optional function that will be called to "tail" the
                step action.

        """
        return [
            cls.from_stack_name(name, context, requires, fn, watch_func)
            for name, requires in graph_dict.items()
        ]

    def __repr__(self) -> str:
        """Object represented as a string."""
        return f"<CFNgin.plan.Step:{self.stack.name}>"

    def __str__(self) -> str:
        """Object displayed as a string."""
        return self.stack.name


class Graph:
    """Graph represents a graph of steps.

    The :class:`Graph` helps organize the steps needed to execute a particular
    action for a set of :class:`runway.cfngin.stack.Stack` objects. When
    initialized with a set of steps, it will first build a Directed Acyclic
    Graph from the steps and their dependencies.

    Example:
        >>> dag = DAG()
        >>> a = Step("a", fn=deploy)
        >>> b = Step("b", fn=deploy)
        >>> dag.add_step(a)
        >>> dag.add_step(b)
        >>> dag.connect(a, b)

    """

    dag: DAG
    steps: Dict[str, Step]

    def __init__(
        self, steps: Optional[Dict[str, Step]] = None, dag: Optional[DAG] = None
    ) -> None:
        """Instantiate class.

        Args:
            steps: Dict with key of step name and value of :class:`Step` for
                steps to initialize the Graph with. Note that if this is provided,
                a pre-configured :class:`runway.cfngin.dag.DAG` that already
                includes these steps should also be provided..
            dag: An optional :class:`runway.cfngin.dag.DAG` object. If one is
                not provided, a new one will be initialized.

        """
        self.steps = steps or {}
        self.dag = dag or DAG()

    def add_step(
        self, step: Step, add_dependencies: bool = False, add_dependents: bool = False
    ) -> None:
        """Add a step to the graph.

        Args:
            step: The step to be added.
            add_dependencies: Connect steps that need to be completed before this
                step.
            add_dependents: Connect steps that require this step.

        """
        self.steps[step.name] = step
        self.dag.add_node(step.name)

        if add_dependencies:
            for dep in step.requires:
                self.connect(step.name, dep)

        if add_dependents:
            for parent in step.required_by:
                self.connect(parent, step.name)

    def add_step_if_not_exists(
        self, step: Step, add_dependencies: bool = False, add_dependents: bool = False
    ) -> None:
        """Try to add a step to the graph.

        Can be used when failure to add is acceptable.

        Args:
            step: The step to be added.
            add_dependencies: Connect steps that need to be completed before this
                step.
            add_dependents: Connect steps that require this step.

        """
        if self.steps.get(step.name):
            return

        self.steps[step.name] = step
        self.dag.add_node_if_not_exists(step.name)

        if add_dependencies:
            for dep in step.requires:
                try:
                    self.connect(step.name, dep)
                except GraphError:
                    continue

        if add_dependents:
            for parent in step.required_by:
                try:
                    self.connect(parent, step.name)
                except GraphError:
                    continue

    def add_steps(self, steps: List[Step]) -> None:
        """Add a list of steps.

        Args:
            steps: The step to be added.

        """
        for step in steps:
            self.add_step(step)

        for step in steps:
            for dep in step.requires:
                self.connect(step.name, dep)

            for parent in step.required_by:
                self.connect(parent, step.name)

    def pop(self, step: Step, default: Any = None) -> Any:
        """Remove a step from the graph.

        Args:
            step: The step to remove from the graph.
            default: Returned if the step could not be popped

        """
        self.dag.delete_node_if_exists(step.name)
        return self.steps.pop(step.name, default)

    def connect(self, step: str, dep: str) -> None:
        """Connect a dependency to a step.

        Args:
            step: Step name to add a dependency to.
            dep: Name of dependent step.

        """
        try:
            self.dag.add_edge(step, dep)
        except (DAGValidationError, KeyError) as exc:
            raise GraphError(exc, step, dep) from None

    def transitive_reduction(self) -> None:
        """Perform a transitive reduction on the underlying DAG.

        The transitive reduction of a graph is a graph with as few edges as
        possible with the same reachability as the original graph.

        See https://en.wikipedia.org/wiki/Transitive_reduction

        """
        self.dag.transitive_reduction()

    def walk(
        self,
        walker: Callable[[DAG, Callable[[str], Any]], Any],
        walk_func: Callable[[Step], Any],
    ) -> Any:
        """Walk the steps of the graph.

        Args:
            walker: Function used to walk the steps.
            walk_func: Function called with a :class:`Step` as the only argument
                for each step of the plan.

        """

        def fn(step_name: str) -> Any:
            """Get a step by step name and execute the ``walk_func`` on it.

            Args:
                step_name: Name of a step.

            """
            step = self.steps[step_name]
            return walk_func(step)

        return walker(self.dag, fn)

    def downstream(self, step_name: str) -> List[Step]:
        """Return the direct dependencies of the given step."""
        return [self.steps[dep] for dep in self.dag.downstream(step_name)]

    def transposed(self) -> Graph:
        """Return a "transposed" version of this graph.

        Useful for walking in reverse.

        """
        return Graph(steps=self.steps, dag=self.dag.transpose())

    def filtered(self, step_names: List[str]) -> Graph:
        """Return a "filtered" version of this graph.

        Args:
            step_names: Steps to filter.

        """
        return Graph(steps=self.steps, dag=self.dag.filter(step_names))

    def topological_sort(self) -> List[Step]:
        """Perform a topological sort of the underlying DAG."""
        nodes = self.dag.topological_sort()
        return [self.steps[step_name] for step_name in nodes]

    def to_dict(self) -> OrderedDict[str, Set[str]]:
        """Return the underlying DAG as a dictionary."""
        return self.dag.graph

    def dumps(self, indent: Optional[int] = None) -> str:
        """Output the graph as a json serialized string for storage.

        Args:
            indent: Number of spaces for each indentation.

        """
        return json.dumps(self.to_dict(), default=json_serial, indent=indent)

    @classmethod
    def from_dict(
        cls,
        graph_dict: Union[
            Dict[str, List[str]], Dict[str, Set[str]], OrderedDict[str, Set[str]]
        ],
        context: CfnginContext,
    ) -> Graph:
        """Create a Graph from a graph dict.

        Args:
            graph_dict: The dictionary used to create the graph.
            context: Required to init stacks.

        """
        return cls.from_steps(Step.from_persistent_graph(graph_dict, context))

    @classmethod
    def from_steps(cls, steps: List[Step]) -> Graph:
        """Create a Graph from Steps.

        Args:
            steps: Steps used to create the graph.

        """
        graph = cls()
        graph.add_steps(steps)
        return graph

    def __str__(self) -> str:
        """Object displayed as a string."""
        return self.dumps()


class Plan:
    """A convenience class for working on a Graph.

    Attributes:
        context: Context object.
        description: Plan description.
        graph: Graph of the plan.
        id: UUID for the plan.
        reverse: The graph has been transposed for walking in reverse.
        require_unlocked: Require the persistent graph to be unlocked before
            executing steps.

    """

    context: Optional[CfnginContext]
    description: str
    graph: Graph
    id: uuid.UUID
    require_unlocked: bool
    reverse: bool

    def __init__(
        self,
        description: str,
        graph: Graph,
        context: Optional[CfnginContext] = None,
        reverse: bool = False,
        require_unlocked: bool = True,
    ) -> None:
        """Initialize class.

        Args:
            description: Description of what the plan is going to do.
            graph: Local graph used for the plan.
            context: Context object.
            reverse: Transpose the graph for walking in reverse.
            require_unlocked: Require the persistent graph to be
                unlocked before executing steps.

        """
        self.context = context
        self.description = description
        self.id = uuid.uuid4()
        self.reverse = reverse
        self.require_unlocked = require_unlocked

        if self.reverse:
            graph = graph.transposed()

        if self.context:
            self.locked = self.context.persistent_graph_locked

            if self.context.stack_names:
                nodes = [
                    target
                    for target in self.context.stack_names
                    if graph.steps.get(target)
                ]

                graph = graph.filtered(nodes)
        else:
            self.locked = False

        self.graph = graph

    def outline(self, level: int = logging.INFO, message: str = ""):
        """Print an outline of the actions the plan is going to take.

        The outline will represent the rough ordering of the steps that will be
        taken.

        Args:
            level: a valid log level that should be used to log
                the outline
            message: a message that will be logged to
                the user after the outline has been logged.

        """
        LOGGER.log(level, 'plan "%s":', self.description)
        for steps, step in enumerate(self.steps, start=1):
            LOGGER.log(
                level,
                '  - step: %s: target: "%s", action: "%s"',
                steps,
                step.name,
                step.fn.__name__ if callable(step.fn) else step.fn,
            )
        if message:
            LOGGER.log(level, message)

    def dump(
        self,
        *,
        directory: str,
        context: CfnginContext,
        provider: Optional[Provider] = None,
    ) -> Any:
        """Output the rendered blueprint for all stacks in the plan.

        Args:
            directory: Directory where files will be created.
            context: Current CFNgin context.
            provider: Provider to use when resolving the blueprints.

        """
        LOGGER.info('dumping "%s"...', self.description)
        directory = os.path.expanduser(directory)
        if not os.path.exists(directory):
            os.makedirs(directory)

        def walk_func(step: Step) -> bool:
            """Walk function."""
            step.stack.resolve(context=context, provider=provider)
            blueprint = step.stack.blueprint
            filename = stack_template_key_name(blueprint)
            path = os.path.join(directory, filename)

            blueprint_dir = os.path.dirname(path)
            if not os.path.exists(blueprint_dir):
                os.makedirs(blueprint_dir)

            LOGGER.info('writing stack "%s" -> %s', step.name, path)
            with open(path, "w", encoding="utf-8") as _file:
                _file.write(blueprint.rendered)

            return True

        return self.graph.walk(walk, walk_func)

    def execute(self, *args: Any, **kwargs: Any):
        """Walk each step in the underlying graph.

        Raises:
            PersistentGraphLocked: Raised if the persistent graph is
                locked prior to execution and this session did not lock it.
            PlanFailed: Raised if any of the steps fail.

        """
        if self.locked and self.require_unlocked:
            raise PersistentGraphLocked
        self.walk(*args, **kwargs)

        failed_steps = [step for step in self.steps if step.status == FAILED]
        if failed_steps:
            raise PlanFailed(failed_steps)

    def walk(self, walker: Callable[..., Any]) -> Any:
        """Walk each step in the underlying graph, in topological order.

        Args:
            walker: a walker function to be passed to :class:`runway.cfngin.dag.DAG`
                to walk the graph.

        """

        def walk_func(step: Step) -> bool:
            """Execute a :class:`Step` wile walking the graph.

            Handles updating the persistent graph if one is being used.

            Args:
                step: :class:`Step` to execute.

            """
            # Before we execute the step, we need to ensure that it's
            # transitive dependencies are all in an "ok" state. If not, we
            # won't execute this step.
            for dep in self.graph.downstream(step.name):
                if not dep.ok:
                    step.set_status(FailedStatus("dependency has failed"))
                    return step.ok

            result = step.run()

            if not self.context or not self.context.persistent_graph:
                return result

            if step.completed or (
                step.skipped
                and step.status.reason == ("does not exist in cloudformation")
            ):
                fn_name = step.fn.__name__ if callable(step.fn) else step.fn
                if fn_name == "_destroy_stack":
                    self.context.persistent_graph.pop(step)
                    LOGGER.debug(
                        "removed step '%s' from the persistent graph", step.name
                    )
                elif fn_name == "_launch_stack":
                    self.context.persistent_graph.add_step_if_not_exists(
                        step, add_dependencies=True, add_dependents=True
                    )
                    LOGGER.debug("added step '%s' to the persistent graph", step.name)
                else:
                    return result
                self.context.put_persistent_graph(self.lock_code)
            return result

        return self.graph.walk(walker, walk_func)

    @property
    def lock_code(self) -> str:
        """Code to lock/unlock the persistent graph."""
        return str(self.id)

    @property
    def steps(self) -> List[Step]:
        """Return a list of all steps in the plan."""
        steps = self.graph.topological_sort()
        steps.reverse()
        return steps

    @property
    def step_names(self) -> List[str]:
        """Return a list of all step names."""
        return [step.name for step in self.steps]

    def keys(self) -> List[str]:
        """Return a list of all step names."""
        return self.step_names



"""Protocols for structural typing.

For more information on protocols, refer to
`PEP 544 <https://www.python.org/dev/peps/pep-0544/>`__.

"""
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Optional, TypeVar, Union, overload

from typing_extensions import Protocol, runtime_checkable

if TYPE_CHECKING:
    from ...context import CfnginContext


_T = TypeVar("_T")


class CfnginHookArgsProtocol(Protocol):
    """Protocol for CFNgin hook arguments class.

    This class defines a structural interface for all CFNgin hook argument
    classes. It is recommended to use the provided base class in place of this
    when authoring a new argument class.

    """

    @overload
    @abstractmethod
    def get(self, __name: str) -> Optional[Any]:
        ...

    @overload
    @abstractmethod
    def get(self, __name: str, __default: Union[Any, _T]) -> Union[Any, _T]:
        ...

    @abstractmethod
    def get(self, __name: str, __default: Union[Any, _T] = None) -> Union[Any, _T]:
        """Safely get the value of an attribute.

        Args:
            name: Attribute name to return the value for.
            default: Value to return if attribute is not found.

        """
        raise NotImplementedError

    @abstractmethod
    def __contains__(self, __name: str) -> bool:
        raise NotImplementedError

    @abstractmethod
    def __getattribute__(self, __name: str) -> Any:
        raise NotImplementedError

    @abstractmethod
    def __getitem__(self, __name: str) -> Any:
        raise NotImplementedError

    @abstractmethod
    def __setitem__(self, __name: str, _value: Any) -> None:
        raise NotImplementedError


@runtime_checkable
class CfnginHookProtocol(Protocol):
    """Protocol for CFNgin hooks.

    This class defines a structural interface for all CFNgin hook classes.
    Classes used for hooks do not need to subclass this hook. They only need to
    implement a similar interface. While not required, it is still acceptable
    to subclass this class for full type checking of a hook class.

    """

    args: CfnginHookArgsProtocol
    """Arguments passed to the hook and parsed into an object."""

    @abstractmethod
    def __init__(  # pylint: disable=super-init-not-called
        self, context: CfnginContext, **_kwargs: Any
    ) -> None:
        """Structural __init__ method.

        This should not be called. Pylint will erroneously warn about
        "super-init-not-called" if using this class as a subclass. This should
        be disabled in-line until the bug reports for this issue is resolved.

        """
        raise NotImplementedError

    @abstractmethod
    def post_deploy(self) -> Any:
        """Run during the **post_deploy** stage.

        Returns:
            A "truthy" value if the hook was successful or a "falsy" value if the
            hook failed.

        """
        raise NotImplementedError

    @abstractmethod
    def post_destroy(self) -> Any:
        """Run during the **post_destroy** stage.

        Returns:
            A "truthy" value if the hook was successful or a "falsy" value if the
            hook failed.

        """
        raise NotImplementedError

    @abstractmethod
    def pre_deploy(self) -> Any:
        """Run during the **pre_deploy** stage.

        Returns:
            A "truthy" value if the hook was successful or a "falsy" value if the
            hook failed.

        """
        raise NotImplementedError

    @abstractmethod
    def pre_destroy(self) -> Any:
        """Run during the **pre_destroy** stage.

        Returns:
            A "truthy" value if the hook was successful or a "falsy" value if the
            hook failed.

        """
        raise NotImplementedError



#!/usr/bin/env python
"""Module with k8s nodegroup."""
import json
import os
from typing import Any

import botocore
from troposphere import Base64, Equals, If, Not, NoValue, Output, Sub, autoscaling, ec2
from troposphere.policies import AutoScalingRollingUpdate, UpdatePolicy

from runway.cfngin.blueprints.base import Blueprint
from runway.cfngin.blueprints.variables.types import (
    CFNNumber,
    CFNString,
    EC2ImageId,
    EC2SecurityGroupId,
    EC2SubnetIdList,
    EC2VPCId,
)


def get_valid_instance_types() -> Any:
    """Return list of instance types."""
    ec2_service_file = os.path.join(
        os.path.dirname(botocore.__file__),
        "data",
        "ec2",
        "2016-11-15",
        "service-2.json",
    )
    # This encoding needs to be explicitly called out as utf-8 on Windows
    # (or it will try cp1252 instead)
    with open(ec2_service_file, "r", encoding="utf-8") as stream:
        return json.load(stream)["shapes"]["InstanceType"]["enum"]


class NodeGroup(Blueprint):
    """CFNgin blueprint for creating k8s nodegroup."""

    VARIABLES = {
        "KeyName": {
            "type": CFNString,  # string to allow it to be unset
            "description": "(Optional) EC2 Key Pair to allow SSH "
            "access to the instances",
            "default": "",
        },
        "NodeImageId": {
            "type": EC2ImageId,
            "description": "AMI id for the node instances.",
        },
        "NodeInstanceType": {
            "type": CFNString,
            "description": "EC2 instance type for the node " "instances",
            "default": "t2.medium",
            "allowed_values": get_valid_instance_types(),
            "constraint_description": "Must be a valid EC2 " "instance type",
        },
        "NodeInstanceProfile": {
            "type": CFNString,
            "description": "Instance profile for the nodes.",
        },
        "NodeAutoScalingGroupMinSize": {
            "type": CFNNumber,
            "description": "Minimum size of Node " "Group ASG.",
            "default": 1,
        },
        "NodeAutoScalingGroupMaxSize": {
            "type": CFNNumber,
            "description": "Maximum size of Node " "Group ASG.",
            "default": 3,
        },
        "NodeVolumeSize": {
            "type": CFNNumber,
            "description": "Node volume size",
            "default": 20,
        },
        "ClusterName": {
            "type": CFNString,
            "description": "The cluster name provided when the "
            "cluster was created. If it is "
            "incorrect, nodes will not be able to "
            "join the cluster.",
        },
        "BootstrapArguments": {
            "type": CFNString,
            "description": "Arguments to pass to the "
            "bootstrap script. See "
            "files/bootstrap.sh in "
            "https://github.com/awslabs/amazon-eks-ami",
            "default": "",
        },
        "NodeGroupName": {
            "type": CFNString,
            "description": "Unique identifier for the Node " "Group.",
        },
        "ClusterControlPlaneSecurityGroup": {
            "type": EC2SecurityGroupId,
            "description": "The security " "group of the " "cluster control " "plane.",
        },
        "VpcId": {"type": EC2VPCId, "description": "The VPC of the worker instances"},
        "Subnets": {
            "type": EC2SubnetIdList,
            "description": "The subnets where workers can be " "created.",
        },
        "UseDesiredInstanceCount": {
            "type": CFNString,
            "description": "Should the initial bootstrap instance count be used?",
        },
    }

    def create_template(self) -> None:
        """Create template (main function called by CFNgin)."""
        template = self.template
        template.set_version("2010-09-09")
        template.set_description(
            "Kubernetes workers via EKS - V1.0.0 "
            "- compatible with amazon-eks-node-v23+"
        )

        # Metadata
        template.set_metadata(
            {
                "AWS::CloudFormation::Interface": {
                    "ParameterGroups": [
                        {
                            "Label": {"default": "EKS Cluster"},
                            "Parameters": [
                                self.variables[i].name
                                for i in [
                                    "ClusterName",
                                    "ClusterControlPlaneSecurityGroup",
                                ]
                            ],
                        },
                        {
                            "Label": {"default": "Worker Node Configuration"},
                            "Parameters": [
                                self.variables[i].name
                                for i in [
                                    "NodeGroupName",
                                    "NodeAutoScalingGroupMinSize",
                                    "NodeAutoScalingGroupMaxSize",
                                    "UseDesiredInstanceCount",
                                    "NodeInstanceType",
                                    "NodeInstanceProfile",
                                    "NodeImageId",
                                    "NodeVolumeSize",
                                    "KeyName",
                                    "BootstrapArguments",
                                ]
                            ],
                        },
                        {
                            "Label": {"default": "Worker Network Configuration"},
                            "Parameters": [
                                self.variables[i].name for i in ["VpcId", "Subnets"]
                            ],
                        },
                    ]
                }
            }
        )

        # Conditions
        template.add_condition(
            "DesiredInstanceCountSpecified",
            Equals(self.variables["UseDesiredInstanceCount"].ref, "true"),
        )
        template.add_condition(
            "KeyNameSpecified", Not(Equals(self.variables["KeyName"].ref, ""))
        )

        # Resources
        nodesecuritygroup = template.add_resource(
            ec2.SecurityGroup(
                "NodeSecurityGroup",
                GroupDescription="Security group for all nodes in the cluster",
                Tags=[
                    {
                        "Key": Sub("kubernetes.io/cluster/${ClusterName}"),
                        "Value": "owned",
                    },
                ],
                VpcId=self.variables["VpcId"].ref,
            )
        )
        template.add_output(
            Output(
                "NodeSecurityGroup",
                Description="Security group for all nodes in the cluster",
                Value=nodesecuritygroup.ref(),
            )
        )
        template.add_resource(
            ec2.SecurityGroupIngress(
                "NodeSecurityGroupIngress",
                Description="Allow node to communicate with each other",
                GroupId=nodesecuritygroup.ref(),
                SourceSecurityGroupId=nodesecuritygroup.ref(),
                IpProtocol="-1",
                FromPort=0,
                ToPort=65535,
            )
        )
        template.add_resource(
            ec2.SecurityGroupIngress(
                "NodeSecurityGroupFromControlPlaneIngress",
                Description="Allow worker Kubelets and pods to receive "
                "communication from the cluster control plane",
                GroupId=nodesecuritygroup.ref(),
                SourceSecurityGroupId=self.variables[
                    "ClusterControlPlaneSecurityGroup"
                ].ref,
                IpProtocol="tcp",
                FromPort=1025,
                ToPort=65535,
            )
        )
        template.add_resource(
            ec2.SecurityGroupEgress(
                "ControlPlaneEgressToNodeSecurityGroup",
                Description="Allow the cluster control plane to communicate "
                "with worker Kubelet and pods",
                GroupId=self.variables["ClusterControlPlaneSecurityGroup"].ref,
                DestinationSecurityGroupId=nodesecuritygroup.ref(),
                IpProtocol="tcp",
                FromPort=1025,
                ToPort=65535,
            )
        )
        template.add_resource(
            ec2.SecurityGroupIngress(
                "NodeSecurityGroupFromControlPlaneOn443Ingress",
                Description="Allow pods running extension API servers on port "
                "443 to receive communication from cluster "
                "control plane",
                GroupId=nodesecuritygroup.ref(),
                SourceSecurityGroupId=self.variables[
                    "ClusterControlPlaneSecurityGroup"
                ].ref,  # noqa
                IpProtocol="tcp",
                FromPort=443,
                ToPort=443,
            )
        )
        template.add_resource(
            ec2.SecurityGroupEgress(
                "ControlPlaneEgressToNodeSecurityGroupOn443",
                Description="Allow the cluster control plane to communicate "
                "with pods running extension API servers on port "
                "443",
                GroupId=self.variables["ClusterControlPlaneSecurityGroup"].ref,
                DestinationSecurityGroupId=nodesecuritygroup.ref(),
                IpProtocol="tcp",
                FromPort=443,
                ToPort=443,
            )
        )
        template.add_resource(
            ec2.SecurityGroupIngress(
                "ClusterControlPlaneSecurityGroupIngress",
                Description="Allow pods to communicate with the cluster API " "Server",
                GroupId=self.variables["ClusterControlPlaneSecurityGroup"].ref,
                SourceSecurityGroupId=nodesecuritygroup.ref(),
                IpProtocol="tcp",
                FromPort=443,
                ToPort=443,
            )
        )

        nodelaunchtemplate = template.add_resource(
            ec2.LaunchTemplate(
                "NodeLaunchTemplate",
                LaunchTemplateData=ec2.LaunchTemplateData(
                    BlockDeviceMappings=[
                        ec2.LaunchTemplateBlockDeviceMapping(
                            DeviceName="/dev/xvda",
                            Ebs=ec2.EBSBlockDevice(
                                DeleteOnTermination=True,
                                VolumeSize=self.variables["NodeVolumeSize"].ref,
                                VolumeType="gp2",
                            ),
                        ),
                    ],
                    IamInstanceProfile=ec2.IamInstanceProfile(
                        Arn=self.variables["NodeInstanceProfile"].ref
                    ),
                    ImageId=self.variables["NodeImageId"].ref,
                    InstanceType=self.variables["NodeInstanceType"].ref,
                    KeyName=If(
                        "KeyNameSpecified", self.variables["KeyName"].ref, NoValue
                    ),
                    MetadataOptions=ec2.MetadataOptions(
                        HttpPutResponseHopLimit=2,
                        HttpEndpoint="enabled",
                        HttpTokens="optional",
                    ),
                    SecurityGroupIds=[nodesecuritygroup.ref()],
                    UserData=Base64(
                        Sub(
                            "\n".join(
                                [
                                    "#!/bin/bash",
                                    "set -o xtrace",
                                    "/etc/eks/bootstrap.sh ${ClusterName} ${BootstrapArguments}",
                                    "/opt/aws/bin/cfn-signal --exit-code $? \\",
                                    "         --stack  ${AWS::StackName} \\",
                                    "         --resource NodeGroup  \\",
                                    "         --region ${AWS::Region}",
                                    "sudo yum install -y "
                                    "https://s3.amazonaws.com/ec2-downloads-windows/SSMAgent/"
                                    "latest/linux_amd64/amazon-ssm-agent.rpm",
                                    "sudo systemctl enable amazon-ssm-agent",
                                    "sudo systemctl start amazon-ssm-agent",
                                ]
                            )
                        )
                    ),
                ),
            )
        )

        template.add_resource(
            autoscaling.AutoScalingGroup(
                "NodeGroup",
                DesiredCapacity=If(
                    "DesiredInstanceCountSpecified",
                    self.variables["NodeAutoScalingGroupMaxSize"].ref,
                    NoValue,
                ),
                LaunchTemplate=autoscaling.LaunchTemplateSpecification(
                    LaunchTemplateId=nodelaunchtemplate.ref(),
                    Version=nodelaunchtemplate.get_att("LatestVersionNumber"),
                ),
                MinSize=self.variables["NodeAutoScalingGroupMinSize"].ref,
                MaxSize=self.variables["NodeAutoScalingGroupMaxSize"].ref,
                Tags=[
                    autoscaling.Tag(
                        "Name", Sub("${ClusterName}-${NodeGroupName}-Node"), True
                    ),
                    autoscaling.Tag(
                        Sub("kubernetes.io/cluster/${ClusterName}"), "owned", True
                    ),
                ],
                VPCZoneIdentifier=self.variables["Subnets"].ref,
                UpdatePolicy=UpdatePolicy(
                    AutoScalingRollingUpdate=AutoScalingRollingUpdate(
                        MinInstancesInService="1", MaxBatchSize="1"
                    )
                ),
            )
        )


# Helper section to enable easy blueprint -> template generation
# (just run `python <thisfile>` to output the json)
if __name__ == "__main__":
    from runway.context import CfnginContext

    print(  # noqa: T201
        NodeGroup("test", CfnginContext(parameters={"namespace": "test"})).to_json()
    )



"""Python dependency compatibility handling."""
import sys
from typing import Iterable

if sys.version_info < (3, 8):  # 3.7
    import shlex

    from backports.cached_property import cached_property
    from importlib_metadata import PackageNotFoundError, version

    def shlex_join(split_command: Iterable[str]) -> str:
        """Backport of :meth:`shlex.join`."""
        return " ".join(shlex.quote(arg) for arg in split_command)

else:
    from functools import cached_property
    from importlib.metadata import PackageNotFoundError, version
    from shlex import join as shlex_join

__all__ = [
    "PackageNotFoundError",
    "cached_property",
    "shlex_join",
    "version",
]



"""Blueprint for the Authorization@Edge implementation of a Static Site.

Described in detail in this blogpost:
https://aws.amazon.com/blogs/networking-and-content-delivery/authorizationedge-how-to-use-lambdaedge-and-json-web-tokens-to-enhance-web-application-security/

"""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Optional

import awacs.logs
import awacs.s3
from awacs.aws import Allow, Principal, Statement
from troposphere import Join, Output, awslambda, cloudfront, iam, s3

from .staticsite import StaticSite

if TYPE_CHECKING:
    from ...cfngin.blueprints.type_defs import BlueprintVariableTypeDef
    from ...context import CfnginContext

LOGGER = logging.getLogger("runway")


class AuthAtEdge(StaticSite):
    """Auth@Edge Blueprint."""

    AUTH_VARIABLES: Dict[str, BlueprintVariableTypeDef] = {
        "OAuthScopes": {"type": list, "default": [], "description": "OAuth2 Scopes"},
        "PriceClass": {
            "type": str,
            "default": "PriceClass_100",  # US/Europe
            "description": "CF price class for the distribution.",
        },
        "RedirectPathSignIn": {
            "type": str,
            "default": "/parseauth",
            "description": "Auth@Edge: The URL that should "
            "handle the redirect from Cognito "
            "after sign-in.",
        },
        "RedirectPathAuthRefresh": {
            "type": str,
            "default": "/refreshauth",
            "description": "The URL path that should "
            "handle the JWT refresh request.",
        },
        "NonSPAMode": {
            "type": bool,
            "default": False,
            "description": "Whether Auth@Edge should omit SPA specific settings",
        },
        "SignOutUrl": {
            "type": str,
            "default": "/signout",
            "description": "The URL path that you can visit to sign-out.",
        },
    }
    IAM_ARN_PREFIX = "arn:aws:iam::aws:policy/service-role/"
    VARIABLES: ClassVar[Dict[str, BlueprintVariableTypeDef]] = {}

    def __init__(
        self,
        name: str,
        context: CfnginContext,
        mappings: Optional[Dict[str, Dict[str, Any]]] = None,
        description: Optional[str] = None,
    ) -> None:
        """Initialize the Blueprint.

        Args:
            name: A name for the blueprint.
            context: Context the blueprint is being executed under.
            mappings: CloudFormation Mappings to be used in the template.
            description: Used to describe the resulting CloudFormation template.

        """
        super().__init__(
            name=name, context=context, description=description, mappings=mappings
        )
        self.VARIABLES.update(StaticSite.VARIABLES)
        self.VARIABLES.update(self.AUTH_VARIABLES)

    def create_template(self) -> None:
        """Create the Blueprinted template for Auth@Edge."""
        self.template.set_version("2010-09-09")
        self.template.set_description(
            "Authorization@Edge Static Website - Bucket, Lambdas, and Distribution"
        )

        # Resources
        bucket = self.add_bucket()
        oai = self.add_origin_access_identity()
        bucket_policy = self.add_cloudfront_bucket_policy(bucket, oai)
        # TODO Make this available in Auth@Edge
        lambda_function_associations: List[cloudfront.LambdaFunctionAssociation] = []

        if self.directory_index_specified:
            index_rewrite = self._get_index_rewrite_role_function_and_version()
            lambda_function_associations = self.get_directory_index_lambda_association(
                lambda_function_associations, index_rewrite["version"]
            )

        # Auth@Edge Lambdas
        check_auth_name = "CheckAuth"
        check_auth_lambda = self.get_auth_at_edge_lambda_and_ver(
            check_auth_name,
            "Check Authorization information for request",
            "check_auth",
            self.add_lambda_execution_role(
                "CheckAuthLambdaExecutionRole", check_auth_name
            ),
        )
        http_headers_name = "HttpHeaders"
        http_headers_lambda = self.get_auth_at_edge_lambda_and_ver(
            http_headers_name,
            "Additional Headers added to every response",
            "http_headers",
            self.add_lambda_execution_role(
                "HttpHeadersLambdaExecutionRole", http_headers_name
            ),
        )
        parse_auth_name = "ParseAuth"
        parse_auth_lambda = self.get_auth_at_edge_lambda_and_ver(
            parse_auth_name,
            "Parse the Authorization Headers/Cookies for the request",
            "parse_auth",
            self.add_lambda_execution_role(
                "ParseAuthLambdaExecutionRole", parse_auth_name
            ),
        )
        refresh_auth_name = "RefreshAuth"
        refresh_auth_lambda = self.get_auth_at_edge_lambda_and_ver(
            refresh_auth_name,
            "Refresh the Authorization information when expired",
            "refresh_auth",
            self.add_lambda_execution_role(
                "RefreshAuthLambdaExecutionRole", refresh_auth_name
            ),
        )
        sign_out_name = "SignOut"
        sign_out_lambda = self.get_auth_at_edge_lambda_and_ver(
            sign_out_name,
            "Sign the User out of the application",
            "sign_out",
            self.add_lambda_execution_role("SignOutLambdaExecutionRole", sign_out_name),
        )

        # CloudFront Distribution
        distribution_options = self.get_distribution_options(
            bucket,
            oai,
            lambda_function_associations,
            check_auth_lambda["version"],
            http_headers_lambda["version"],
            parse_auth_lambda["version"],
            refresh_auth_lambda["version"],
            sign_out_lambda["version"],
        )
        self.add_cloudfront_distribution(bucket_policy, distribution_options)

    def get_auth_at_edge_lambda_and_ver(
        self, title: str, description: str, handle: str, role: iam.Role
    ) -> Dict[str, Any]:
        """Create a lambda function and its version.

        Args:
            title: The name of the function in PascalCase.
            description: Description to be displayed in the
                lambda panel.
            handle: The underscore separated representation
                of the name of the lambda. This handle is used to
                determine the handler for the lambda as well as
                identify the correct Code hook_data information.
            role: The Lambda Execution Role.

        """
        function = self.get_auth_at_edge_lambda(title, description, handle, role)
        return {"function": function, "version": self.add_version(title, function)}

    def get_auth_at_edge_lambda(
        self, title: str, description: str, handler: str, role: iam.Role
    ) -> awslambda.Function:
        """Create an Auth@Edge lambda resource.

        Args:
            title: The name of the function in PascalCase.
            description: Description to be displayed in the lambda panel.
            handler: The underscore separated representation
                of the name of the lambda. This handle is used to
                determine the handler for the lambda as well as
                identify the correct Code hook_data information.
            role: The Lambda Execution Role.

        """
        lamb = self.template.add_resource(
            awslambda.Function(
                title,
                DeletionPolicy="Retain",
                Code=self.context.hook_data["aae_lambda_config"][handler],
                Description=description,
                Handler="__init__.handler",
                Role=role.get_att("Arn"),
                Runtime="python3.7",
            )
        )

        self.template.add_output(
            Output(
                f"Lambda{title}Arn",
                Description=f"Arn For the {title} Lambda Function",
                Value=lamb.get_att("Arn"),
            )
        )

        return lamb

    def add_version(
        self, title: str, lambda_function: awslambda.Function
    ) -> awslambda.Version:
        """Create a version association with a Lambda@Edge function.

        In order to ensure different versions of the function
        are appropriately uploaded a hash based on the code of the
        lambda is appended to the name. As the code changes so
        will this hash value.

        Args:
            title: The name of the function in PascalCase.
            lambda_function: The Lambda function.

        """
        s3_key = lambda_function.properties["Code"].to_dict()["S3Key"]
        code_hash = s3_key.split(".")[0].split("-")[-1]
        return self.template.add_resource(
            awslambda.Version(
                title + "Ver" + code_hash, FunctionName=lambda_function.ref()
            )
        )

    def get_distribution_options(
        self,
        bucket: s3.Bucket,
        oai: cloudfront.CloudFrontOriginAccessIdentity,
        lambda_funcs: List[cloudfront.LambdaFunctionAssociation],
        check_auth_lambda_version: awslambda.Version,
        http_headers_lambda_version: awslambda.Version,
        parse_auth_lambda_version: awslambda.Version,
        refresh_auth_lambda_version: awslambda.Version,
        sign_out_lambda_version: awslambda.Version,
    ) -> Dict[str, Any]:
        """Retrieve the options for our CloudFront distribution.

        Keyword Args:
            bucket: The bucket resource.
            oai: The origin access identity resource.
            lambda_funcs: List of Lambda Function associations.
            check_auth_lambda_version: Lambda Function Version to use.
            http_headers_lambda_version: Lambda Function Version to use.
            parse_auth_lambda_version: Lambda Function Version to use.
            refresh_auth_lambda_version: Lambda Function Version to use.
            sign_out_lambda_version: Lambda Function Version to use.

        Return:
            The CloudFront Distribution Options.

        """
        default_cache_behavior_lambdas = lambda_funcs
        default_cache_behavior_lambdas.append(
            cloudfront.LambdaFunctionAssociation(
                EventType="viewer-request",
                LambdaFunctionARN=check_auth_lambda_version.ref(),
            )
        )
        default_cache_behavior_lambdas.append(
            cloudfront.LambdaFunctionAssociation(
                EventType="origin-response",
                LambdaFunctionARN=http_headers_lambda_version.ref(),
            )
        )

        return {
            "Aliases": self.add_aliases(),
            "Origins": [
                cloudfront.Origin(
                    DomainName=Join(".", [bucket.ref(), "s3.amazonaws.com"]),
                    S3OriginConfig=cloudfront.S3OriginConfig(
                        OriginAccessIdentity=Join(
                            "", ["origin-access-identity/cloudfront/", oai.ref()]
                        )
                    ),
                    Id="protected-bucket",
                )
            ],
            "CacheBehaviors": [
                cloudfront.CacheBehavior(
                    PathPattern=self.variables["RedirectPathSignIn"],
                    Compress=True,
                    ForwardedValues=cloudfront.ForwardedValues(QueryString=True),
                    LambdaFunctionAssociations=[
                        cloudfront.LambdaFunctionAssociation(
                            EventType="viewer-request",
                            LambdaFunctionARN=parse_auth_lambda_version.ref(),
                        )
                    ],
                    TargetOriginId="protected-bucket",
                    ViewerProtocolPolicy="redirect-to-https",
                ),
                cloudfront.CacheBehavior(
                    PathPattern=self.variables["RedirectPathAuthRefresh"],
                    Compress=True,
                    ForwardedValues=cloudfront.ForwardedValues(QueryString=True),
                    LambdaFunctionAssociations=[
                        cloudfront.LambdaFunctionAssociation(
                            EventType="viewer-request",
                            LambdaFunctionARN=refresh_auth_lambda_version.ref(),
                        )
                    ],
                    TargetOriginId="protected-bucket",
                    ViewerProtocolPolicy="redirect-to-https",
                ),
                cloudfront.CacheBehavior(
                    PathPattern=self.variables["SignOutUrl"],
                    Compress=True,
                    ForwardedValues=cloudfront.ForwardedValues(QueryString=True),
                    LambdaFunctionAssociations=[
                        cloudfront.LambdaFunctionAssociation(
                            EventType="viewer-request",
                            LambdaFunctionARN=sign_out_lambda_version.ref(),
                        )
                    ],
                    TargetOriginId="protected-bucket",
                    ViewerProtocolPolicy="redirect-to-https",
                ),
            ],
            "DefaultCacheBehavior": cloudfront.DefaultCacheBehavior(
                AllowedMethods=["GET", "HEAD"],
                Compress=self.variables.get("Compress", True),
                DefaultTTL="86400",
                ForwardedValues=cloudfront.ForwardedValues(QueryString=True),
                LambdaFunctionAssociations=default_cache_behavior_lambdas,
                TargetOriginId="protected-bucket",
                ViewerProtocolPolicy="redirect-to-https",
            ),
            "DefaultRootObject": "index.html",
            "Logging": self.add_logging_bucket(),
            "PriceClass": self.variables["PriceClass"],
            "Enabled": True,
            "WebACLId": self.add_web_acl(),
            "CustomErrorResponses": self._get_error_responses(),
            "ViewerCertificate": self.add_acm_cert(),
        }

    def _get_error_responses(self) -> List[cloudfront.CustomErrorResponse]:
        """Return error response based on site stack variables.

        When custom_error_responses are defined return those, if running
        in NonSPAMode return nothing, or return the standard error responses
        for a SPA.

        """
        if self.variables["custom_error_responses"]:
            return [
                cloudfront.CustomErrorResponse(
                    ErrorCode=response["ErrorCode"],
                    ResponseCode=response["ResponseCode"],
                    ResponsePagePath=response["ResponsePagePath"],
                )
                for response in self.variables["custom_error_responses"]
            ]
        if self.variables["NonSPAMode"]:
            return []
        return [
            cloudfront.CustomErrorResponse(
                ErrorCode=404, ResponseCode=200, ResponsePagePath="/index.html"
            )
        ]

    # pyright: reportIncompatibleMethodOverride=none
    def _get_cloudfront_bucket_policy_statements(  # pylint: disable=arguments-differ
        self, bucket: s3.Bucket, oai: cloudfront.CloudFrontOriginAccessIdentity
    ) -> List[Statement]:
        return [
            Statement(
                Action=[awacs.s3.GetObject],
                Effect=Allow,
                Principal=Principal("CanonicalUser", oai.get_att("S3CanonicalUserId")),
                Resource=[Join("", [bucket.get_att("Arn"), "/*"])],
            ),
            Statement(
                Action=[awacs.s3.ListBucket],
                Effect=Allow,
                Principal=Principal("CanonicalUser", oai.get_att("S3CanonicalUserId")),
                Resource=[bucket.get_att("Arn")],
            ),
        ]



"""K8s (kustomize) module."""
from __future__ import annotations

import logging
import subprocess
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union, cast

from typing_extensions import Literal

from .._logging import PrefixAdaptor
from ..compat import cached_property
from ..config.models.runway.options.k8s import RunwayK8sModuleOptionsDataModel
from ..core.components import DeployEnvironment
from ..env_mgr.kbenv import KBEnvManager
from ..exceptions import KubectlVersionNotSpecified
from ..utils import which
from .base import ModuleOptions, RunwayModule
from .utils import run_module_command

if TYPE_CHECKING:
    from .._logging import RunwayLogger
    from ..context import RunwayContext

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))

KubectlCommandTypeDef = Literal[
    "annotation",
    "apply",
    "auth",
    "autoscale",
    "cp",
    "create",
    "delete",
    "describe",
    "diff",
    "edit",
    "exec",
    "expose",
    "get",
    "kustomize",
    "label",
    "logs",
    "patch",
    "port-forward",
    "proxy",
]


class K8s(RunwayModule):
    """Kubectl Runway Module."""

    DEPRECATION_MSG = (
        "Kubectl Runway module support has been deprecated and "
        "may be removed in the next major release."
    )

    options: K8sOptions

    def __init__(
        self,
        context: RunwayContext,
        *,
        explicitly_enabled: Optional[bool] = False,
        logger: RunwayLogger = LOGGER,
        module_root: Path,
        name: Optional[str] = None,
        options: Optional[Union[Dict[str, Any], ModuleOptions]] = None,
        parameters: Optional[Dict[str, Any]] = None,
        **_: Any,
    ) -> None:
        """Instantiate class.

        Args:
            context: Runway context object for the current session.
            explicitly_enabled: Whether or not the module is explicitly enabled.
                This is can be set in the event that the current environment being
                deployed to matches the defined environments of the module/deployment.
            logger: Used to write logs.
            module_root: Root path of the module.
            name: Name of the module.
            options: Options passed to the module class from the config as ``options``
                or ``module_options`` if coming from the deployment level.
            parameters: Values to pass to the underlying infrastructure as code
                tool that will alter the resulting infrastructure being deployed.
                Used to templatize IaC.

        """
        super().__init__(
            context,
            explicitly_enabled=explicitly_enabled,
            logger=logger,
            module_root=module_root,
            name=name,
            options=K8sOptions.parse_obj(
                deploy_environment=context.env, obj=options or {}, path=module_root
            ),
            parameters=parameters,
        )
        # logger needs to be created here to use the correct logger
        self.logger = PrefixAdaptor(self.name, LOGGER)
        LOGGER.warning("%s:%s", self.name, self.DEPRECATION_MSG)

    @cached_property
    def kbenv(self) -> KBEnvManager:
        """Kubectl environment manager."""
        return KBEnvManager(self.path, overlay_path=self.options.overlay_path)

    @cached_property
    def kubectl_bin(self) -> str:
        """Path to kubectl binary."""
        try:
            return self.kbenv.install(self.options.kubectl_version)
        except KubectlVersionNotSpecified as exc:
            self.logger.verbose("kubectl version not specified; checking path")
            if not which("kubectl"):
                self.logger.error(
                    "kubectl not available and a version to install not specified"
                )
                self.logger.error(exc.message)
                sys.exit(1)
            return "kubectl"

    @cached_property
    def skip(self) -> bool:
        """Determine if the module should be skipped."""
        if self.options.kustomize_config.is_file():
            LOGGER.info(
                "processing kustomize overlay: %s", self.options.kustomize_config
            )
            return False
        LOGGER.info(
            "skipped; kustomize overlay for this environment/region not"
            " found -- looking for one of: %s",
            ", ".join(
                str(self.path / "overlays" / i / "kustomization.yaml")
                for i in self.options.gen_overlay_dirs(
                    self.ctx.env.name, self.ctx.env.aws_region
                )
            ),
        )
        return True

    def deploy(self) -> None:
        """Run kubectl apply."""
        if self.skip:
            return
        self.kubectl_kustomize()
        self.kubectl_apply()

    def destroy(self) -> None:
        """Run kubectl delete."""
        if self.skip:
            return
        self.kubectl_kustomize()
        self.kubectl_delete()

    def gen_cmd(
        self,
        command: KubectlCommandTypeDef,
        args_list: Optional[List[str]] = None,
    ) -> List[str]:
        """Generate and log a kubectl command.

        This does not execute the command, only prepares it for use.

        Args:
            command: The CDK command to be executed.
            args_list: Additional arguments to include in the generated command.

        Returns:
            The full command to be passed into a subprocess.

        """
        cmd_list = [self.kubectl_bin, command]
        cmd_list.extend(args_list or [])
        if command in ["apply", "delete"]:
            cmd_list.extend(["--kustomize", str(self.options.overlay_path)])
            if command == "delete":
                cmd_list.append("--ignore-not-found=true")
        elif command == "kustomize":
            cmd_list.append(str(self.options.overlay_path))
        LOGGER.debug("running command: %s", " ".join(cmd_list))
        return cmd_list

    def init(self) -> None:
        """Run init."""
        LOGGER.warning("init not currently supported for %s", self.__class__.__name__)

    def kubectl_apply(self) -> None:
        """Execute ``kubectl apply`` command.

        https://kubectl.docs.kubernetes.io/references/kubectl/apply/

        """
        self.logger.info("deploy (in progress)")
        run_module_command(
            cmd_list=self.gen_cmd("apply"),
            env_vars=self.ctx.env.vars,
            logger=self.logger,
        )
        self.logger.info("deploy (complete)")

    def kubectl_delete(self) -> None:
        """Execute ``kubectl delete`` command.

        https://kubectl.docs.kubernetes.io/references/kubectl/delete/

        """
        self.logger.info("destroy (in progress)")
        run_module_command(
            cmd_list=self.gen_cmd("delete"),
            env_vars=self.ctx.env.vars,
            logger=self.logger,
        )
        self.logger.info("destroy (complete)")

    def kubectl_kustomize(self) -> str:
        """Execute ``kubectl kustomize`` command.

        https://kubectl.docs.kubernetes.io/references/kubectl/kustomize/

        """
        kustomize_yml = subprocess.check_output(
            self.gen_cmd("kustomize"), env=self.ctx.env.vars
        ).decode()
        self.logger.debug("kustomized yaml generated by kubectl:\n\n%s", kustomize_yml)
        return kustomize_yml

    def plan(self) -> None:
        """Run kustomize build and display generated plan."""
        if self.skip:
            return
        self.logger.info(
            "kustomized yaml generated by kubectl:\n\n%s", self.kubectl_kustomize()
        )


class K8sOptions(ModuleOptions):
    """Module options for Kubernetes.

    Attributes:
        data: Options parsed into a data model.
        deploy_environment: Runway deploy environment object.
        kubectl_version: Version of kubectl to use.
        path: Module path.

    """

    data: RunwayK8sModuleOptionsDataModel
    deploy_environment: DeployEnvironment
    kubectl_version: Optional[str]
    path: Path

    def __init__(
        self,
        data: RunwayK8sModuleOptionsDataModel,
        deploy_environment: DeployEnvironment,
        path: Path,
    ) -> None:
        """Instantiate class.

        Args:
            data: Options parsed into a data model.
            deploy_environment: Current deploy environment.
            path: Module path.

        """
        self.data = data
        self.env = deploy_environment
        self.kubectl_version = data.kubectl_version
        self.path = path

    @cached_property
    def kustomize_config(self) -> Path:
        """Kustomize configuration file."""
        return self.overlay_path / "kustomization.yaml"

    @cached_property
    def overlay_path(self) -> Path:
        """Directory containing the kustomize overlay to use."""
        if self.data.overlay_path:
            return self.data.overlay_path
        return self.get_overlay_dir(
            path=self.path / "overlays",
            environment=self.env.name,
            region=self.env.aws_region,
        )

    @staticmethod
    def gen_overlay_dirs(environment: str, region: str) -> List[str]:
        """Generate possible overlay directories.

        Prefers more explicit directory name but falls back to environment name only.

        Args:
            environment: Current deploy environment.
            region : Current AWS region.

        """
        return [f"{environment}-{region}", environment]

    @classmethod
    def get_overlay_dir(cls, path: Path, environment: str, region: str) -> Path:
        """Determine the overlay directory to use."""
        overlay_dir = path
        for name in cls.gen_overlay_dirs(environment, region):
            overlay_dir = path / name
            if (overlay_dir / "kustomization.yaml").is_file():
                return overlay_dir
        return overlay_dir

    @classmethod
    def parse_obj(
        cls,
        deploy_environment: DeployEnvironment,
        obj: object,
        path: Optional[Path] = None,
    ) -> K8sOptions:
        """Parse options definition and return an options object.

        Args:
            deploy_environment: Current deploy environment.
            obj: Object to parse.
            path: Module path.

        """
        return cls(
            data=RunwayK8sModuleOptionsDataModel.parse_obj(obj),
            deploy_environment=deploy_environment,
            path=path or Path.cwd(),
        )



"""Retrieve a value from an environment variable."""
# pyright: reportIncompatibleMethodOverride=none
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Union

from typing_extensions import Final, Literal

from .base import LookupHandler

if TYPE_CHECKING:
    from ...context import CfnginContext, RunwayContext


class EnvLookup(LookupHandler):
    """Environment variable Lookup."""

    TYPE_NAME: Final[Literal["env"]] = "env"
    """Name that the Lookup is registered as."""

    @classmethod
    def handle(  # pylint: disable=arguments-differ
        cls,
        value: str,
        context: Union[CfnginContext, RunwayContext],
        *__args: Any,
        **__kwargs: Any,
    ) -> Any:
        """Retrieve an environment variable.

        The value is retrieved from a copy of the current environment variables
        that is saved to the context object. These environment variables
        are manipulated at runtime by Runway to fill in additional values
        such as ``DEPLOY_ENVIRONMENT`` and ``AWS_REGION`` to match the
        current execution.

        Args:
            value: The value passed to the Lookup.
            context: The current context object.

        Raises:
            ValueError: Unable to find a value for the provided query and
                a default value was not provided.

        """
        query, args = cls.parse(value)

        result = context.env.vars.get(query, args.pop("default", ""))

        if result != "":  # allows for False bool and NoneType results
            return cls.format_results(result, **args)

        raise ValueError(f'"{value}" does not exist in the environment')



"""CFNgin hook for syncing static website to S3 bucket."""
from __future__ import annotations

import hashlib
import json
import logging
import os
import time
from operator import itemgetter
from typing import TYPE_CHECKING, Any, Dict, List, Optional, cast

import yaml

from ....core.providers.aws.s3 import Bucket
from ....module.staticsite.options.models import RunwayStaticSiteExtraFileDataModel
from ....utils import JsonEncoder
from ..base import HookArgsBaseModel

if TYPE_CHECKING:
    from boto3.session import Session

    from ....context import CfnginContext

LOGGER = logging.getLogger(__name__)


class HookArgs(HookArgsBaseModel):
    """Hook arguments."""

    bucket_name: str
    """S3 bucket name."""

    cf_disabled: bool = False
    """Disable the use of CloudFront."""

    distribution_domain: str = "undefined"
    """Domain of the CloudFront distribution."""

    distribution_id: str = "undefined"
    """CloudFront distribution ID."""

    distribution_path: str = "/*"
    """Path in the CloudFront distribution to invalidate."""

    extra_files: List[RunwayStaticSiteExtraFileDataModel] = []
    """Extra files to sync to the S3 bucket."""

    website_url: Optional[str] = None
    """S3 bucket website URL."""


def get_archives_to_prune(
    archives: List[Dict[str, Any]], hook_data: Dict[str, Any]
) -> List[str]:
    """Return list of keys to delete.

    Args:
        archives: The full list of file archives
        hook_data: CFNgin hook data

    """
    files_to_skip = [
        hook_data[i]
        for i in ["current_archive_filename", "old_archive_filename"]
        if hook_data.get(i)
    ]

    archives.sort(  # sort from oldest to newest
        key=itemgetter("LastModified"), reverse=False
    )

    # Drop all but last 15 files
    return [i["Key"] for i in archives[:-15] if i["Key"] not in files_to_skip]


def sync(context: CfnginContext, *__args: Any, **kwargs: Any) -> bool:
    """Sync static website to S3 bucket.

    Arguments parsed by :class:`~runway.cfngin.hooks.staticsite.upload_staticsite.HookArgs`.

    Args:
        context: The context instance.

    """
    args = HookArgs.parse_obj(kwargs)
    session = context.get_session()
    build_context = context.hook_data["staticsite"]
    invalidate_cache = False

    synced_extra_files = sync_extra_files(
        context,
        args.bucket_name,
        args.extra_files,
        hash_tracking_parameter=build_context.get("hash_tracking_parameter"),
    )

    if synced_extra_files:
        invalidate_cache = True

    if build_context["deploy_is_current"]:
        LOGGER.info("skipped upload; latest version already deployed")
    else:
        bucket = Bucket(context, args.bucket_name)
        bucket.sync_from_local(
            build_context["app_directory"],
            delete=True,
            exclude=[f.name for f in args.extra_files if f.name],
        )
        invalidate_cache = True

    if args.cf_disabled:
        LOGGER.info("STATIC WEBSITE URL: %s", args.website_url)
    elif invalidate_cache:
        invalidate_distribution(
            session,
            identifier=args.distribution_id,
            domain=args.distribution_domain,
            path=args.distribution_path,
        )

    LOGGER.info("sync complete")

    if not build_context["deploy_is_current"]:
        update_ssm_hash(context, session)

    prune_archives(context, session)

    return True


def update_ssm_hash(context: CfnginContext, session: Session) -> bool:
    """Update the SSM hash with the new tracking data.

    Args:
        context: Context instance.
        session: boto3 session.

    """
    build_context = context.hook_data["staticsite"]

    if not build_context.get("hash_tracking_disabled"):
        hash_param = build_context["hash_tracking_parameter"]
        hash_value = build_context["hash"]

        LOGGER.info("updating SSM parameter %s with hash %s", hash_param, hash_value)

        set_ssm_value(
            session,
            hash_param,
            hash_value,
            "Hash of currently deployed static website source",
        )

    return True


def invalidate_distribution(
    session: Session,
    *,
    domain: str = "undefined",
    identifier: str,
    path: str = "/*",
    **_: Any,
) -> bool:
    """Invalidate the current distribution.

    Args:
        session: The current CFNgin session.
        domain: The distribution domain.
        identifier: The distribution id.
        path: The distribution path.

    """
    LOGGER.info("invalidating CloudFront distribution: %s (%s)", identifier, domain)
    cf_client = session.client("cloudfront")
    cf_client.create_invalidation(
        DistributionId=identifier,
        InvalidationBatch={
            "Paths": {"Quantity": 1, "Items": [path]},
            "CallerReference": str(time.time()),
        },
    )

    LOGGER.info("CloudFront invalidation complete")
    return True


def prune_archives(context: CfnginContext, session: Session) -> bool:
    """Prune the archives from the bucket.

    Args:
        context: The context instance.
        session: The CFNgin session.

    """
    LOGGER.info("cleaning up old site archives...")
    archives: List[Dict[str, Any]] = []
    s3_client = session.client("s3")
    list_objects_v2_paginator = s3_client.get_paginator("list_objects_v2")
    response_iterator = list_objects_v2_paginator.paginate(
        Bucket=context.hook_data["staticsite"]["artifact_bucket_name"],
        Prefix=context.hook_data["staticsite"]["artifact_key_prefix"],
    )

    for page in response_iterator:
        archives.extend(page.get("Contents", []))  # type: ignore
    archives_to_prune = get_archives_to_prune(archives, context.hook_data["staticsite"])

    # Iterate in chunks of 1000 to match delete_objects limit
    for objects in [
        archives_to_prune[i : i + 1000] for i in range(0, len(archives_to_prune), 1000)
    ]:
        s3_client.delete_objects(
            Bucket=context.hook_data["staticsite"]["artifact_bucket_name"],
            Delete={"Objects": [{"Key": i} for i in objects]},
        )
    return True


def auto_detect_content_type(filename: Optional[str]) -> Optional[str]:
    """Auto detects the content type based on the filename.

    Args:
        filename : A filename to use to auto detect the content type.

    Returns:
        The content type of the file. None if the content type could not be detected.

    """
    if not filename:
        return None

    _, ext = os.path.splitext(filename)

    if ext == ".json":
        return "application/json"

    if ext in [".yml", ".yaml"]:
        return "text/yaml"

    return None


def get_content_type(extra_file: RunwayStaticSiteExtraFileDataModel) -> Optional[str]:
    """Return the content type of the file.

    Args:
        extra_file: The extra file configuration.

    Returns:
        The content type of the extra file. If 'content_type' is provided then
        that is returned, otherwise it is auto detected based on the name.

    """
    return extra_file.content_type or auto_detect_content_type(extra_file.name)


def get_content(extra_file: RunwayStaticSiteExtraFileDataModel) -> Optional[str]:
    """Get serialized content based on content_type.

    Args:
        extra_file: The extra file configuration.

    Returns:
        Serialized content based on the content_type.

    """
    if extra_file.content:
        if isinstance(extra_file.content, (dict, list)):
            if extra_file.content_type == "application/json":
                return json.dumps(extra_file.content)

            if extra_file.content_type == "text/yaml":
                return yaml.safe_dump(extra_file.content)

            raise ValueError(
                '"content_type" must be json or yaml if "content" is not a string'
            )

        if not isinstance(extra_file.content, str):
            raise TypeError(f"unsupported content: {type(extra_file.content)}")

    return cast(Optional[str], extra_file.content)


def calculate_hash_of_extra_files(
    extra_files: List[RunwayStaticSiteExtraFileDataModel],
) -> str:
    """Return a hash of all of the given extra files.

    All attributes of the extra file object are included when hashing:
    name, content_type, content, and file data.

    Args:
        extra_files: The list of extra file configurations.

    Returns:
        The hash of all the files.

    """
    file_hash = hashlib.md5()

    for extra_file in sorted(extra_files, key=lambda x: x.name):
        file_hash.update((extra_file.name + "\0").encode())

        if extra_file.content_type:
            file_hash.update((extra_file.content_type + "\0").encode())

        if extra_file.content:
            LOGGER.debug("hashing content: %s", extra_file.name)
            file_hash.update((cast(str, extra_file.content) + "\0").encode())

        if extra_file.file:
            with open(extra_file.file, "rb") as f:
                LOGGER.debug("hashing file: %s", extra_file.file)
                for chunk in iter(
                    lambda: f.read(4096), ""  # pylint: disable=cell-var-from-loop
                ):
                    if not chunk:
                        break
                    file_hash.update(chunk)
                file_hash.update("\0".encode())

    return file_hash.hexdigest()


def get_ssm_value(session: Session, name: str) -> Optional[str]:
    """Get the ssm parameter value.

    Args:
        session: The boto3 session.
        name: The parameter name.

    Returns:
        The parameter value.

    """
    ssm_client = session.client("ssm")

    try:
        return ssm_client.get_parameter(Name=name)["Parameter"]["Value"]
    except ssm_client.exceptions.ParameterNotFound:
        return None


def set_ssm_value(
    session: Session, name: str, value: Any, description: str = ""
) -> None:
    """Set the ssm parameter.

    Args:
        session: The boto3 session.
        name: The name of the parameter.
        value: The value of the parameter.
        description: A description of the parameter.

    """
    ssm_client = session.client("ssm")

    ssm_client.put_parameter(
        Name=name, Description=description, Value=value, Type="String", Overwrite=True
    )


def sync_extra_files(
    context: CfnginContext,
    bucket: str,
    extra_files: List[RunwayStaticSiteExtraFileDataModel],
    **kwargs: Any,
) -> List[str]:
    """Sync static website extra files to S3 bucket.

    Args:
        context: The context instance.
        bucket: The static site bucket name.
        extra_files: List of files and file content that should be uploaded.

    """
    LOGGER.debug("extra_files to sync: %s", json.dumps(extra_files, cls=JsonEncoder))

    if not extra_files:
        return []

    session = context.get_session()
    s3_client = session.client("s3")
    uploaded: List[str] = []

    hash_param = cast(str, kwargs.get("hash_tracking_parameter", ""))
    hash_new = None

    # serialize content based on content type
    for extra_file in extra_files:
        extra_file.content_type = get_content_type(extra_file)
        extra_file.content = get_content(extra_file)

    # calculate a hash of the extra_files
    if hash_param:
        hash_param = f"{hash_param}extra"

        hash_old = get_ssm_value(session, hash_param)

        # calculate hash of content
        hash_new = calculate_hash_of_extra_files(extra_files)

        if hash_new == hash_old:
            LOGGER.info(
                "skipped upload of extra files; latest version already deployed"
            )
            return []

    for extra_file in extra_files:
        if extra_file.content:
            LOGGER.info("uploading extra file: %s", extra_file.name)

            s3_client.put_object(
                Bucket=bucket,
                Key=extra_file.name,
                Body=str(extra_file.content).encode(),
                ContentType=cast(str, extra_file.content_type),
            )

            uploaded.append(extra_file.name)

        if extra_file.file:
            LOGGER.info(
                "uploading extra file: %s as %s ", extra_file.file, extra_file.name
            )

            extra_args = ""

            if extra_file.content_type:
                extra_args = {"ContentType": extra_file.content_type}

            if extra_args:
                s3_client.upload_file(
                    Bucket=bucket,
                    ExtraArgs=extra_args,
                    Filename=str(extra_file.file),
                    Key=extra_file.name,
                )
            if not extra_args:
                s3_client.upload_file(
                    Bucket=bucket,
                    Filename=str(extra_file.file),
                    Key=extra_file.name,
                )

            uploaded.append(extra_file.name)

    if hash_new:
        LOGGER.info(
            "updating extra files SSM parameter %s with hash %s", hash_param, hash_new
        )
        set_ssm_value(session, hash_param, hash_new)

    return uploaded



"""CFNgin context."""
from __future__ import annotations

import collections.abc
import json
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, MutableMapping, Optional, Union, cast

from pydantic import BaseModel

from .._logging import PrefixAdaptor, RunwayLogger
from ..cfngin.exceptions import (
    PersistentGraphCannotLock,
    PersistentGraphCannotUnlock,
    PersistentGraphLockCodeMismatch,
    PersistentGraphLocked,
    PersistentGraphUnlocked,
)
from ..cfngin.plan import Graph
from ..cfngin.stack import Stack
from ..cfngin.utils import ensure_s3_bucket
from ..compat import cached_property
from ..config import CfnginConfig
from ..core.components import DeployEnvironment
from ._base import BaseContext

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client

    from .type_defs import PersistentGraphLocation

LOGGER = cast(RunwayLogger, logging.getLogger(__name__))


def get_fqn(base_fqn: str, delimiter: str, name: Optional[str] = None) -> str:
    """Return the fully qualified name of an object within this context.

    If the name passed already appears to be a fully qualified name, it
    will be returned with no further processing.

    """
    if name and name.startswith(f"{base_fqn}{delimiter}"):
        return name

    return delimiter.join([_f for _f in [base_fqn, name] if _f])


class CfnginContext(BaseContext):
    """CFNgin context object.

    Attributes:
        bucket_region: Region where the S3 Bucket is located. The S3 Bucket
            being the Bucket configured for staging CloudFormation Templates.
        config: CFNgin configuration file that has been resolved & parsed into a
            python object.
        config_path: Path to the configuration file that has been resolved, parsed
            and made accessible via this object.
        env: Deploy environment object containing information about the current
            deploy environment.
        force_stacks: List of stacks to force.
        hook_data: Values returned by hooks that are stored based on the ``data_key``
            defined for the hook. Returned values are only stored if a ``data_key``
            was provided AND the return value is a dict or ``pydantic.BaseModel``.
        logger: Custom logger to use when logging messages.
        parameters: Parameters passed from Runway or read from a file.
        stack_names: List of Stack names to operate on. If value is falsy, all
            Stacks defined in the config will be operated on.

    """

    _persistent_graph_lock_code: Optional[str]
    _persistent_graph_lock_tag: str = "cfngin_lock_code"
    _persistent_graph: Optional[Graph]
    _s3_bucket_verified: bool

    bucket_region: str
    config: CfnginConfig
    config_path: Path
    env: DeployEnvironment
    force_stacks: List[str]
    hook_data: Dict[str, Any]
    logger: Union[PrefixAdaptor, RunwayLogger]
    parameters: MutableMapping[str, Any]
    stack_names: List[str]

    def __init__(
        self,
        *,
        config: Optional[CfnginConfig] = None,
        config_path: Optional[Path] = None,
        deploy_environment: Optional[DeployEnvironment] = None,
        force_stacks: Optional[List[str]] = None,
        logger: Union[PrefixAdaptor, RunwayLogger] = LOGGER,
        parameters: Optional[MutableMapping[str, Any]] = None,
        stack_names: Optional[List[str]] = None,
        work_dir: Optional[Path] = None,
        **_: Any,
    ) -> None:
        """Instantiate class.

        Args:
            config: The CFNgin configuration being operated on.
            config_path: Path to the config file that was provided.
            deploy_environment: The current deploy environment.
            force_stacks: A list of stacks to force work on. Used to work on locked stacks.
            logger: Custom logger.
            parameters: Parameters passed from Runway or read from a file.
            stack_names: A list of stack_names to operate on. If not passed,
                all stacks defined in the config will be operated on.
            work_dir: Working directory used by Runway.

        """
        self.config_path = config_path or Path.cwd()
        super().__init__(
            deploy_environment=deploy_environment or DeployEnvironment(),
            logger=PrefixAdaptor(
                self.config_path.name.split(".")[0],
                logger if isinstance(logger, RunwayLogger) else LOGGER,
            ),
            work_dir=work_dir or self.config_path / ".runway",
        )
        self._persistent_graph_lock_code = None
        self._persistent_graph = None
        self._s3_bucket_verified = False
        self.config = config or CfnginConfig.parse_obj({"namespace": "example"})
        self.bucket_region = self.config.cfngin_bucket_region or self.env.aws_region
        self.parameters = parameters or {}
        self.force_stacks = force_stacks or []
        self.hook_data = {}
        self.stack_names = stack_names or []

    @cached_property
    def base_fqn(self) -> str:
        """Return ``namespace`` sanitized for use as an S3 Bucket name."""
        return self.config.namespace.replace(".", "-").lower()

    @cached_property
    def bucket_name(self) -> Optional[str]:
        """Return ``cfngin_bucket`` from config, calculated name, or None."""
        if not self.upload_to_s3:
            return None
        return (
            self.config.cfngin_bucket
            or f"cfngin-{self.get_fqn()}-{self.env.aws_region}"
        )

    @cached_property
    def mappings(self) -> Dict[str, Dict[str, Dict[str, Any]]]:
        """Return ``mappings`` from config."""
        return self.config.mappings or {}

    @cached_property
    def namespace(self) -> str:
        """Return ``namespace`` from config."""
        return self.config.namespace

    @cached_property
    def namespace_delimiter(self) -> str:
        """Return ``namespace_delimiter`` from config or default."""
        return self.config.namespace_delimiter

    @cached_property
    def persistent_graph_location(self) -> PersistentGraphLocation:
        """Location of the persistent graph in s3."""
        if not self.bucket_name or not self.config.persistent_graph_key:
            return {}

        key = (
            self.config.persistent_graph_key + ".json"
            if not self.config.persistent_graph_key.endswith(".json")
            else self.config.persistent_graph_key
        )
        return {
            "Bucket": self.bucket_name,
            "Key": f"persistent_graphs/{self.config.namespace}/{key}",
        }

    @property
    def persistent_graph_locked(self) -> bool:
        """Check if persistent graph is locked."""
        if not self.persistent_graph:
            return False
        return bool(self.persistent_graph_lock_code)

    @property
    def persistent_graph_lock_code(self) -> Optional[str]:
        """Code used to lock the persistent graph S3 object."""
        if not self._persistent_graph_lock_code and self.persistent_graph_location:
            self._persistent_graph_lock_code = self.persistent_graph_tags.get(
                self._persistent_graph_lock_tag
            )
        return self._persistent_graph_lock_code

    @property
    def persistent_graph_tags(self) -> Dict[str, str]:
        """Cache of tags on the persistent graph object."""
        try:
            return {
                t["Key"]: t["Value"]
                for t in self.s3_client.get_object_tagging(
                    **self.persistent_graph_location
                ).get("TagSet", [])
            }
        except self.s3_client.exceptions.NoSuchKey:
            self.logger.debug(
                "persistent graph object does not exist in S3; could not get tags"
            )
            return {}

    @property
    def persistent_graph(self) -> Optional[Graph]:
        """Graph if a persistent graph is being used.

        Will create an "empty" object in S3 if one is not found.

        """
        if not self._persistent_graph:
            if not self.persistent_graph_location:
                return None

            content = "{}"

            if self.s3_bucket_verified:
                try:
                    self.logger.debug(
                        "getting persistent graph from s3:\n%s",
                        json.dumps(self.persistent_graph_location, indent=4),
                    )
                    content = (
                        self.s3_client.get_object(
                            ResponseContentType="application/json",
                            **self.persistent_graph_location,
                        )["Body"]
                        .read()
                        .decode("utf-8")
                    )
                except self.s3_client.exceptions.NoSuchKey:
                    self.logger.info(
                        "persistent graph object does not exist in s3; "
                        "creating one now..."
                    )
                    self.s3_client.put_object(
                        Body=content.encode(),
                        ServerSideEncryption="AES256",
                        ACL="bucket-owner-full-control",
                        ContentType="application/json",
                        **self.persistent_graph_location,
                    )
            self.persistent_graph = Graph.from_dict(json.loads(content), self)

        return self._persistent_graph

    @persistent_graph.setter
    def persistent_graph(self, graph: Optional[Graph]) -> None:
        """Load a persistent graph dict as a :class:`runway.cfngin.plan.Graph`."""
        self._persistent_graph = graph

    @property
    def s3_bucket_verified(self) -> bool:
        """Check CFNgin bucket exists and you have access.

        If the CFNgin bucket does not exist, will try to create one.

        """
        if not self._s3_bucket_verified and self.bucket_name:
            ensure_s3_bucket(
                self.s3_client,
                self.bucket_name,
                self.bucket_region,
                create=False,
                persist_graph=bool(self.persistent_graph_location),
            )
            self._s3_bucket_verified = True
        return bool(self._s3_bucket_verified)

    @cached_property
    def s3_client(self) -> S3Client:
        """AWS S3 client."""
        return self.get_session(region=self.bucket_region).client("s3")

    @cached_property
    def stacks_dict(self) -> Dict[str, Stack]:
        """Construct a dict of ``{stack.fqn: Stack}`` for easy access to stacks."""
        return {stack.fqn: stack for stack in self.stacks}

    @cached_property
    def stacks(self) -> List[Stack]:
        """Stacks for the current action."""
        return [
            Stack(
                context=self,
                definition=stack_def,
                enabled=stack_def.enabled,
                force=stack_def.name in self.force_stacks,
                locked=stack_def.locked,
                mappings=self.mappings,
                protected=stack_def.protected,
            )
            for stack_def in self.config.stacks
        ]

    @cached_property
    def tags(self) -> Dict[str, str]:
        """Return ``tags`` from config."""
        return (
            self.config.tags
            if self.config.tags is not None
            else {"cfngin_namespace": self.config.namespace}
            if self.config.namespace
            else {}
        )

    @cached_property
    def template_indent(self) -> int:
        """Return ``template_indent`` from config or default."""
        return self.config.template_indent

    @cached_property
    def upload_to_s3(self) -> bool:
        """Check if S3 should be used for caching/persistent graph."""
        # Don't upload stack templates to S3 if `cfngin_bucket` is
        # explicitly set to an empty string.
        if self.config.cfngin_bucket == "":
            self.logger.debug(
                "not uploading to s3; cfngin_bucket "
                "is explicitly set to an empty string"
            )
            return False

        # If no namespace is specified, and there's no explicit
        # cfngin bucket specified, don't upload to s3. This makes
        # sense because we can't realistically auto generate a cfngin
        # bucket name in this case.
        if not self.config.namespace and not self.config.cfngin_bucket:
            self.logger.debug(
                "not uploading to s3; namespace & cfngin_bucket not provided"
            )
            return False

        return True

    def copy(self) -> CfnginContext:
        """Copy the contents of this object into a new instance."""
        return self.__class__(
            config_path=self.config_path,
            config=self.config,
            deploy_environment=self.env.copy(),
            force_stacks=self.force_stacks,
            logger=self.logger,
            parameters=self.parameters,
            stack_name=self.stack_names,
            work_dir=self.work_dir,
        )

    def get_fqn(self, name: Optional[str] = None) -> str:
        """Return the fully qualified name of an object within this context.

        If the name passed already appears to be a fully qualified name, it
        will be returned with no further processing.

        """
        return get_fqn(self.base_fqn, self.config.namespace_delimiter, name)

    def get_stack(self, name: str) -> Optional[Stack]:
        """Get a stack by name.

        Args:
            name: Name of a Stack as defined in the config.

        """
        stack_fqn = self.get_fqn(name)
        if stack_fqn in self.stacks_dict:  # quickly get stack from fqn
            return self.stacks_dict[stack_fqn]
        # account for stack def using stack_name and stack.name was provided
        for stack in self.stacks:
            if stack.name == name:
                return stack
        return None

    def lock_persistent_graph(self, lock_code: str) -> None:
        """Locks the persistent graph in s3.

        Args:
            lock_code: The code that will be used to lock the S3 object.

        Raises:
            :class:`runway.cfngin.exceptions.PersistentGraphLocked`
            :class:`runway.cfngin.exceptions.PersistentGraphCannotLock`

        """
        if not self.persistent_graph:
            return

        if self.persistent_graph_locked:
            raise PersistentGraphLocked

        try:
            self.s3_client.put_object_tagging(
                Tagging={
                    "TagSet": [
                        {"Key": self._persistent_graph_lock_tag, "Value": lock_code}
                    ]
                },
                **self.persistent_graph_location,
            )
            self.logger.info(
                'locked persistent graph "%s" with lock ID "%s"',
                "/".join(
                    [
                        self.persistent_graph_location["Bucket"],
                        self.persistent_graph_location["Key"],
                    ]
                ),
                lock_code,
            )
        except self.s3_client.exceptions.NoSuchKey as exc:
            raise PersistentGraphCannotLock("s3 object does not exist") from exc

    def put_persistent_graph(self, lock_code: str) -> None:
        """Upload persistent graph to s3.

        Args:
            lock_code (str): The code that will be used to lock the S3 object.

        Raises:
            :class:`runway.cfngin.exceptions.PersistentGraphUnlocked`
            :class:`runway.cfngin.exceptions.PersistentGraphLockCodeMismatch`

        """
        if not self.persistent_graph:
            return

        if not self.persistent_graph.to_dict():
            self.s3_client.delete_object(**self.persistent_graph_location)
            self.logger.debug("removed empty persistent graph object from S3")
            return

        if not self.persistent_graph_locked:
            raise PersistentGraphUnlocked(
                reason="It must be locked by the current session to be updated."
            )

        if self.persistent_graph_lock_code != lock_code:
            raise PersistentGraphLockCodeMismatch(
                lock_code, self.persistent_graph_lock_code
            )

        self.s3_client.put_object(
            Body=self.persistent_graph.dumps(4).encode(),
            ServerSideEncryption="AES256",
            ACL="bucket-owner-full-control",
            ContentType="application/json",
            Tagging=f"{self._persistent_graph_lock_tag}={lock_code}",
            **self.persistent_graph_location,
        )
        self.logger.debug(
            "persistent graph updated:\n%s", self.persistent_graph.dumps(indent=4)
        )

    def set_hook_data(self, key: str, data: Any) -> None:
        """Set hook data for the given key.

        Args:
            key: The key to store the hook data in.
            data: A dictionary of data to store, as returned from a hook.

        """
        if not isinstance(data, (BaseModel, collections.abc.Mapping)):
            raise TypeError(
                f"Hook data (key: {key}) must be an instance of "
                "collections.abc.Mapping (a dictionary for example) or pydantic.BaseModel."
            )

        if key in self.hook_data:
            raise KeyError(
                f"Hook data for key {key} already exists, each hook "
                "must have a unique data_key."
            )

        self.hook_data[key] = data

    def unlock_persistent_graph(self, lock_code: str) -> bool:
        """Unlocks the persistent graph in s3.

        Args:
            lock_code: The code that will be used to lock the S3 object.

        Raises:
            :class:`runway.cfngin.exceptions.PersistentGraphCannotUnlock`

        """
        if not self.persistent_graph:
            return True

        if not self.persistent_graph.to_dict():
            try:
                self.s3_client.get_object(
                    ResponseContentType="application/json",
                    **self.persistent_graph_location,
                )
            except self.s3_client.exceptions.NoSuchKey:
                self.logger.info(
                    "persistent graph deleted; does not need to be unlocked"
                )
                return True

        self.logger.verbose(
            'unlocking persistent graph "%s"...', self.persistent_graph_location
        )

        if not self.persistent_graph_locked:
            raise PersistentGraphCannotUnlock(
                PersistentGraphUnlocked(
                    reason="It must be locked by the current session to be unlocked."
                )
            )

        if self.persistent_graph_lock_code == lock_code:
            try:
                self.s3_client.delete_object_tagging(**self.persistent_graph_location)
            except self.s3_client.exceptions.NoSuchKey:
                pass
            self._persistent_graph_lock_code = None
            self.logger.info(
                'unlocked persistent graph "%s/%s"',
                self.persistent_graph_location.get("Bucket"),
                self.persistent_graph_location.get("Key"),
            )
            return True
        raise PersistentGraphCannotUnlock(
            PersistentGraphLockCodeMismatch(lock_code, self.persistent_graph_lock_code)
        )



"""Serverless module."""
from __future__ import annotations

import argparse
import logging
import os
import re
import subprocess
import sys
import tempfile
import uuid
from pathlib import Path
from typing import IO, TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union, cast

import yaml

from .._logging import PrefixAdaptor
from ..cfngin.hooks.staticsite.utils import get_hash_of_files
from ..compat import cached_property
from ..config.models.runway.options.serverless import (
    RunwayServerlessModuleOptionsDataModel,
)
from ..s3_utils import does_s3_object_exist, download, upload
from ..utils import YamlDumper, merge_dicts
from .base import ModuleOptions, RunwayModuleNpm
from .utils import generate_node_command, run_module_command

if TYPE_CHECKING:
    from .._logging import RunwayLogger
    from ..context import RunwayContext
    from ..type_defs import AnyPath, AnyPathConstrained

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))


def gen_sls_config_files(stage: str, region: str) -> List[str]:
    """Generate possible SLS config files names."""
    names: List[str] = []
    for ext in ["yml", "json"]:
        # Give preference to explicit stage-region files
        names.append(os.path.join("env", f"{stage}-{region}.{ext}"))
        names.append(f"config-{stage}-{region}.{ext}")
        # Fallback to stage name only
        names.append(os.path.join("env", f"{stage}.{ext}"))
        names.append(f"config-{stage}.{ext}")
    return names


class Serverless(RunwayModuleNpm):
    """Serverless Runway Module."""

    options: ServerlessOptions

    def __init__(
        self,
        context: RunwayContext,
        *,
        explicitly_enabled: Optional[bool] = False,
        logger: RunwayLogger = LOGGER,
        module_root: Path,
        name: Optional[str] = None,
        options: Optional[Union[Dict[str, Any], ModuleOptions]] = None,
        parameters: Optional[Dict[str, Any]] = None,
        **_: Any,
    ) -> None:
        """Instantiate class.

        Args:
            context: Runway context object for the current session.
            explicitly_enabled: Whether or not the module is explicitly enabled.
                This is can be set in the event that the current environment being
                deployed to matches the defined environments of the module/deployment.
            logger: Used to write logs.
            module_root: Root path of the module.
            name: Name of the module.
            options: Options passed to the module class from the config as ``options``
                or ``module_options`` if coming from the deployment level.
            parameters: Values to pass to the underlying infrastructure as code
                tool that will alter the resulting infrastructure being deployed.
                Used to templatize IaC.

        """
        super().__init__(
            context,
            explicitly_enabled=explicitly_enabled,
            logger=logger,
            module_root=module_root,
            name=name,
            options=ServerlessOptions.parse_obj(options or {}),
            parameters=parameters,
        )
        self.logger = PrefixAdaptor(self.name, logger)
        self.stage = self.ctx.env.name

    @property
    def cli_args(self) -> List[str]:
        """Generate CLI args from self used in all Serverless commands."""
        result = ["--region", self.region, "--stage", self.stage]
        if "DEBUG" in self.ctx.env.vars:
            result.append("--verbose")
        return result

    @cached_property
    def env_file(self) -> Optional[Path]:
        """Find the environment file for the module."""
        for name in gen_sls_config_files(self.stage, self.region):
            test_path = self.path / name
            if test_path.is_file():
                return test_path
        return None

    @property
    def skip(self) -> bool:
        """Determine if the module should be skipped."""
        if not self.package_json_missing():
            if self.parameters or self.explicitly_enabled or self.env_file:
                return False
            self.logger.info(
                "skipped; config file for this stage/region not found"
                " -- looking for one of: %s",
                ", ".join(gen_sls_config_files(self.stage, self.region)),
            )
        else:
            self.logger.info(
                'skipped; package.json with "serverless" in devDependencies'
                " is required for this module type"
            )
        return True

    def extend_serverless_yml(self, func: Callable[..., None]) -> None:
        """Extend the Serverless config file with additional YAML from options.

        Args:
            func: Callable to use after handling the Serverless config file.

        """
        self.npm_install()  # doing this here for a cleaner log
        self.logger.info("extending Serverless config from runway.yml...")
        final_yml = merge_dicts(
            self.sls_print(skip_install=True), self.options.extend_serverless_yml
        )
        # using a unique name to prevent collisions when run in parallel
        tmp_file = self.path / f"{uuid.uuid4()}.tmp.serverless.yml"

        try:
            tmp_file.write_text(yaml.safe_dump(final_yml))
            self.logger.debug("created temporary Serverless config: %s", tmp_file)
            self.options.update_args("config", str(tmp_file.name))
            self.logger.debug(
                "updated options.args with temporary Serverless config: %s",
                tmp_file.name,
            )
            func(skip_install=True)
        finally:
            try:
                tmp_file.unlink()  # always cleanup the temp file
                self.logger.debug("removed temporary Serverless config")
            except OSError:
                self.logger.debug(
                    "encountered an error when trying to delete the "
                    "temporary Serverless config",
                    exc_info=True,
                )

    def gen_cmd(self, command: str, args_list: Optional[List[str]] = None) -> List[str]:
        """Generate and log a Serverless command.

        This does not execute the command, only prepares it for use.

        Args:
            command: The Serverless command to be executed.
            args_list: Additional arguments to include in the generated command.

        Returns:
            The full command to be passed into a subprocess.

        """
        args = [command] + self.cli_args + self.options.args
        args.extend(args_list or [])
        if command not in ["remove", "package", "print"] and self.ctx.is_noninteractive:
            args.append("--conceal")  # hide secrets from serverless output

        # Serverless is no longer happy using "--no-color" arg.
        # It was never natively supported by the app, only by an
        # upstream dependency called "chalk" that it uses. Later
        # versions of Serverless now break with "--no-color"
        # defined. This update disables Serverless color in a way
        # now recommended by the developers, see the following issue:
        # https://github.com/serverless/serverless/issues/11142
        if self.ctx.no_color:
            return generate_node_command(
                command="FORCE_COLOR=0 sls",
                command_opts=args,
                path=self.path,
                logger=self.logger,
            )

        return generate_node_command(
            command="sls", command_opts=args, path=self.path, logger=self.logger
        )

    def sls_deploy(
        self, *, package: Optional[AnyPath] = None, skip_install: bool = False
    ) -> None:
        """Execute ``sls deploy`` command.

        Args:
            package: Path to Serverless package to deploy.
            skip_install: Skip ``npm ci|install`` before running the
                Serverless command.

        """
        if not skip_install:
            self.npm_install()
        run_module_command(
            cmd_list=self.gen_cmd(
                "deploy", args_list=["--package", str(package)] if package else []
            ),
            env_vars=self.ctx.env.vars,
            logger=self.logger,
        )

    def sls_package(
        self,
        *,
        output_path: Optional[AnyPathConstrained] = None,
        skip_install: bool = False,
    ) -> Optional[AnyPathConstrained]:
        """Execute ``sls package`` command.

        Args:
            output_path: Path where the package should be output.
            skip_install: Skip ``npm ci|install`` before running the
                Serverless command.

        """
        if not skip_install:
            self.npm_install()
        run_module_command(
            cmd_list=self.gen_cmd(
                "package",
                args_list=["--package", str(output_path)] if output_path else [],
            ),
            env_vars=self.ctx.env.vars,
            logger=self.logger,
        )
        return output_path

    def sls_print(
        self, *, item_path: Optional[str] = None, skip_install: bool = False
    ) -> Dict[str, Any]:
        """Execute ``sls print`` command.

        Keyword Args:
            item_path: Period-separated path to print a sub-value (eg: "provider.name").
            skip_install: Skip ``npm ci|install`` before running the Serverless command.

        Returns:
            Resolved Serverless config file.

        Raises:
            SystemExit: If a runway-tmp.serverless.yml file already exists.

        """
        if not skip_install:
            self.npm_install()

        args = ["--format", "yaml"]
        if item_path:
            args.extend(["--path", item_path])
        result = yaml.safe_load(
            subprocess.check_output(
                self.gen_cmd("print", args_list=args),
                # disable all deprecation messages to ensure the output is "clean"
                env={"SLS_DEPRECATION_DISABLE": "*", **self.ctx.env.vars},
            )
        )
        # this could be expensive so only dump if needed
        if self.logger.getEffectiveLevel() == logging.DEBUG:
            self.logger.debug(  # cov: ignore
                "resolved Serverless config:\n%s", yaml.dump(result, Dumper=YamlDumper)
            )
        return result

    def sls_remove(self, *, skip_install: bool = False) -> None:
        """Execute ``sls remove`` command.

        Args:
            skip_install: Skip ``npm ci|install`` before running the Serverless command.

        """
        if not skip_install:
            self.npm_install()
        stack_missing = False  # track output for acceptable error
        self.logger.info("destroy (in progress)")
        with subprocess.Popen(
            self.gen_cmd("remove"),
            bufsize=1,
            env=self.ctx.env.vars,
            stdout=subprocess.PIPE,
            universal_newlines=True,
        ) as proc:
            with cast(IO[str], proc.stdout):
                for line in cast(IO[str], proc.stdout):
                    print(line, end="")  # noqa: T201
                    if re.search(r"Stack '.*' does not exist", line):
                        stack_missing = True
        if proc.wait() != 0 and not stack_missing:
            sys.exit(proc.returncode)
        self.logger.info("destroy (complete)")

    def deploy(self) -> None:
        """Entrypoint for Runway's deploy action."""
        if self.skip:
            return
        if self.options.extend_serverless_yml:
            self.extend_serverless_yml(self._deploy_package)
        else:
            self._deploy_package()

    def _deploy_package(self) -> None:
        """Deploy Serverless package."""
        if self.options.promotezip.bucketname:
            with tempfile.TemporaryDirectory(dir=self.ctx.work_dir) as tmp_dir:
                artifact = ServerlessArtifact(
                    self.ctx,
                    self.sls_print(),
                    logger=self.logger,
                    package_path=tmp_dir,
                    path=self.path,
                )
                self.logger.info("package (in progress)")
                self.sls_package(output_path=artifact.package_path, skip_install=True)
                self.logger.info("package (complete)")
                artifact.sync_with_s3(self.options.promotezip.bucketname)
                self.logger.info("deploy (in progress)")
                self.sls_deploy(package=artifact.package_path, skip_install=True)
                self.logger.info("deploy (complete)")
        else:
            self.logger.info("deploy (in progress)")
            self.sls_deploy()
            self.logger.info("deploy (complete)")

    def destroy(self) -> None:
        """Entrypoint for Runway's destroy action."""
        if self.skip:
            return
        if self.options.extend_serverless_yml:
            self.extend_serverless_yml(self.sls_remove)
        else:
            self.sls_remove()

    def init(self) -> None:
        """Run init."""
        self.logger.warning(
            "init not currently supported for %s", self.__class__.__name__
        )

    def plan(self) -> None:
        """Entrypoint for Runway's plan action."""
        self.logger.info("plan not currently supported for Serverless")


class ServerlessArtifact:
    """Object for interacting with a Serverless artifact directory."""

    def __init__(
        self,
        context: RunwayContext,
        config: Dict[str, Any],
        *,
        logger: Union[PrefixAdaptor, RunwayLogger] = LOGGER,
        package_path: AnyPath,
        path: AnyPath,
    ) -> None:
        """Instantiate class.

        Args:
            context: Runway context object.
            config: Rendered Serverless config file.
            logger: Logger this object will log to. If not provided, the logger
                in the local module will be used.
            package_path: Local path to the artifact directory.
            path: Root directory of the Serverless project.

        """
        self.ctx = context
        self.config = config
        self.logger = logger
        self.package_path = (
            Path(package_path) if isinstance(package_path, str) else package_path
        )
        self.path = Path(path) if isinstance(path, str) else path

    @cached_property
    def source_hash(self) -> Dict[str, str]:
        """File hash(es) of each service's source code."""
        if self.config.get("package", {"": ""}).get("individually"):
            return {
                name: get_hash_of_files(
                    self.path / os.path.dirname(detail.get("handler"))
                )
                for name, detail in self.config.get("functions", {}).items()
            }
        directories: List[Dict[str, Union[List[str], str]]] = []
        for _name, detail in self.config.get("functions", {}).items():
            func_path = {"path": os.path.dirname(detail.get("handler"))}
            if func_path not in directories:
                directories.append(func_path)
        if isinstance(self.config["service"], dict):
            # handle sls<3.0.0 potential service property object notation
            return {
                self.config["service"]["name"]: get_hash_of_files(
                    self.path, directories
                )
            }
        return {self.config["service"]: get_hash_of_files(self.path, directories)}

    def sync_with_s3(self, bucket_name: str) -> None:
        """Sync local archive files with S3 bucket.

        Args:
            bucket_name: Name of S3 bucket to upload files to.

        """
        session = self.ctx.get_session()
        for name, file_hash in self.source_hash.items():
            file_path = self.package_path / f"{name}.zip"
            obj_key = f"{file_hash}.zip"
            if does_s3_object_exist(
                bucket_name,
                obj_key,
                session=session,
                region=self.ctx.env.aws_region,
            ):
                self.logger.info("found existing package for %s", name)
                download(
                    bucket=bucket_name,
                    key=obj_key,
                    file_path=str(file_path),
                    session=session,
                )
            else:
                self.logger.info("no existing package found for %s", name)
                if not file_path.is_file():
                    self.logger.info("local file not found for %s", name)
                    continue
                upload(
                    bucket=bucket_name,
                    key=obj_key,
                    filename=str(file_path),
                    session=session,
                )


class ServerlessOptions(ModuleOptions):
    """Module options for Serverless Framework.

    Attributes:
        data: Options parsed into a data model.
        extend_serverless_yml: If provided, the value of this option will be
            recursively merged into the module's Serverless config file.
        promotezip: If provided, promote Serverless Framework generated zip files
            between environments from a build AWS account.
        skip_npm_ci: Skip running ``npm ci`` in the module directory prior to
            processing the module.

    """

    def __init__(self, data: RunwayServerlessModuleOptionsDataModel) -> None:
        """Instantiate class.

        Args:
            data: Options parsed into a data model.

        """
        self._arg_parser = self._create_arg_parser()
        cli_args, self._unknown_cli_args = self._arg_parser.parse_known_args(
            data.args.copy()
        )
        self._cli_args = vars(cli_args)  # convert argparse.Namespace to dict

        self.data = data
        self.extend_serverless_yml = data.extend_serverless_yml
        self.promotezip = data.promotezip
        self.skip_npm_ci = data.skip_npm_ci

    @property
    def args(self) -> List[str]:
        """List of CLI arguments/options to pass to the Serverless Framework CLI."""
        known_args: List[str] = []
        for key, val in self._cli_args.items():
            if isinstance(val, str):
                known_args.extend([f"--{key}", val])
        return known_args + self._unknown_cli_args

    def update_args(self, key: str, value: str) -> None:
        """Update a known CLI argument.

        Args:
            key: Dict key to be updated.
            value: New value

        Raises:
            KeyError: The key provided for update is not a known arg.

        """
        if key in self._cli_args:
            self._cli_args[key] = value
        else:
            raise KeyError(key)

    @staticmethod
    def _create_arg_parser() -> argparse.ArgumentParser:
        """Create argparse parser to parse args.

        Used to pull arguments out of self.args when logic could change
        depending on values provided.

        Returns:
            argparse.ArgumentParser

        """
        parser = argparse.ArgumentParser()
        parser.add_argument("-c", "--config", default=None)
        return parser

    @classmethod
    def parse_obj(cls, obj: object) -> ServerlessOptions:
        """Parse options definition and return an options object.

        Args:
            obj: Object to parse.

        """
        return cls(data=RunwayServerlessModuleOptionsDataModel.parse_obj(obj))



"""Helper functions for bootstrap actions."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from botocore.exceptions import ClientError

from runway.lookups.handlers.base import LookupHandler
from runway.utils import BaseModel

if TYPE_CHECKING:
    from runway.cfngin.providers.aws.default import Provider
    from runway.context import CfnginContext

TYPE_NAME = "bootstrap_value"


class HookArgs(BaseModel):
    """Hook arguments.

    Attributes:
        bootstrap: Value to return during bootstrap.
        post_bootstrap: Value to return post-bootstrap.

    """

    bootstrap: str
    post_bootstrap: str


class BootstrapValue(LookupHandler):
    """Return the bootstrap value on creation otherwise the post_bootstrap.

    .. rubric:: Example
    .. code-block:: yaml

        variables:
          variable: ${bootstrap_value <stack.name>::bootstrap=true, post_bootstrap=false}

    """

    # pylint: disable=arguments-differ
    @classmethod
    def handle(  # type: ignore
        cls,
        value: str,
        context: CfnginContext,
        *_args: Any,
        provider: Provider,
        **_kwargs: Any,
    ) -> str:
        """Handle lookup."""
        query, raw_args = cls.parse(value)
        args = HookArgs.parse_obj(raw_args)

        stack = context.get_stack(query)
        if not stack:
            raise ValueError(f"stack {query} not defined in CFNgin config")
        try:
            stack_des = provider.cloudformation.describe_stacks(StackName=stack.fqn)[
                "Stacks"
            ][0]
        except ClientError as exc:
            if "does not exist" not in str(exc):
                raise
            return args.bootstrap

        if provider.is_stack_completed(stack_des) or (
            provider.is_stack_in_progress(stack_des)
        ):
            return args.post_bootstrap
        return args.bootstrap



"""CFNgin config models."""
# pylint: disable=no-self-argument
from __future__ import annotations

import copy
import locale
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

import yaml
from pydantic import Extra, Field, Protocol, root_validator, validator
from typing_extensions import Literal

from .. import utils
from ..base import ConfigProperty
from ._package_sources import (
    CfnginPackageSourcesDefinitionModel,
    GitCfnginPackageSourceDefinitionModel,
    LocalCfnginPackageSourceDefinitionModel,
    S3CfnginPackageSourceDefinitionModel,
)

if TYPE_CHECKING:
    from pydantic import BaseModel

    Model = TypeVar("Model", bound=BaseModel)

__all__ = [
    "CfnginConfigDefinitionModel",
    "CfnginHookDefinitionModel",
    "CfnginPackageSourcesDefinitionModel",
    "CfnginStackDefinitionModel",
    "GitCfnginPackageSourceDefinitionModel",
    "LocalCfnginPackageSourceDefinitionModel",
    "S3CfnginPackageSourceDefinitionModel",
]


class CfnginHookDefinitionModel(ConfigProperty):
    """Model for a CFNgin hook definition."""

    args: Dict[str, Any] = Field(
        default={},
        title="Arguments",
        description="Arguments that will be passed to the hook. (supports lookups)",
    )
    data_key: Optional[str] = Field(
        default=None,
        description="Key to use when storing the returned result of the hook.",
    )
    enabled: bool = Field(default=True, description="Whether the hook will be run.")
    path: str = Field(..., description="Python importable path to the hook.")
    required: bool = Field(
        default=True,
        description="Whether to continue execution if the hook results in an error.",
    )

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        schema_extra = {
            "description": "Python classes or functions run before or after "
            "deploy/destroy actions."
        }
        title = "CFNgin Hook Definition"


class CfnginStackDefinitionModel(ConfigProperty):
    """Model for a CFNgin stack definition."""

    class_path: Optional[str] = Field(
        default=None,
        title="Blueprint Class Path",
        description="Python importable path to a blueprint class.",
    )
    description: Optional[str] = Field(
        default=None,
        title="Stack Description",
        description="A description that will be applied to the stack in CloudFormation.",
    )
    enabled: bool = Field(
        default=True, description="Whether the stack will be deployed."
    )
    in_progress_behavior: Optional[Literal["wait"]] = Field(
        default=None,
        title="Stack In Progress Behavior",
        description="The action to take when a stack's status is "
        "CREATE_IN_PROGRESS or UPDATE_IN_PROGRESS when trying to update it.",
    )
    locked: bool = Field(
        default=False, description="Whether to limit updating of the stack."
    )
    name: str = Field(..., title="Stack Name", description="Name of the stack.")
    protected: bool = Field(
        default=False,
        description="Whether to force all updates to the stack to be performed interactively.",
    )
    required_by: List[str] = Field(
        default=[], description="Array of stacks (by name) that require this stack."
    )
    requires: List[str] = Field(
        default=[], description="Array of stacks (by name) that this stack requires."
    )
    stack_name: Optional[str] = Field(
        default=None,
        title="Explicit Stack Name",
        description="Explicit name of the stack (namespace will still be prepended).",
    )
    stack_policy_path: Optional[Path] = Field(
        default=None,
        description="Path to a stack policy document that will be applied to the "
        "CloudFormation stack.",
    )
    tags: Dict[str, Any] = Field(
        default={}, description="Tags that will be applied to the CloudFormation stack."
    )
    template_path: Optional[Path] = Field(
        default=None,
        description="Path to a JSON or YAML formatted CloudFormation Template.",
    )
    termination_protection: bool = Field(
        default=False,
        description="Set the value of termination protection on the CloudFormation stack.",
    )
    timeout: Optional[int] = Field(
        default=None,
        description="The amount of time (in minutes) that can pass before the "
        "Stack status becomes CREATE_FAILED.",
    )
    variables: Dict[str, Any] = Field(
        default={},
        description="Parameter values that will be passed to the "
        "Blueprint/CloudFormation stack. (supports lookups)",
    )

    class Config(ConfigProperty.Config):
        """Model configuration options."""

        extra = Extra.forbid
        title = "CFNgin Stack Definition"

        @staticmethod
        def schema_extra(schema: Dict[str, Any]) -> None:  # type: ignore
            """Process the schema after it has been generated.

            Schema is modified in place. Return value is ignored.

            https://pydantic-docs.helpmanual.io/usage/schema/#schema-customization

            """
            schema[
                "description"
            ] = "Define CloudFormation stacks using a Blueprint or Template."
            # prevents a false error when defining stacks as a dict
            schema.get("required", ["name"]).remove("name")

            # fields that can be bool or lookup
            for prop in ["enabled", "locked", "protected", "termination_protection"]:
                schema["properties"][prop].pop("type")
                schema["properties"][prop]["anyOf"] = [
                    {"type": "boolean"},
                    {"type": "string", "pattern": utils.CFNGIN_LOOKUP_STRING_REGEX},
                ]

    _resolve_path_fields = cast(
        "classmethod[Callable[..., Any]]",
        validator("stack_policy_path", "template_path", allow_reuse=True)(
            utils.resolve_path_field
        ),
    )

    @root_validator(pre=True)
    def _validate_class_and_template(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Validate class_path and template_path are not both provided."""
        if values.get("class_path") and values.get("template_path"):
            raise ValueError("only one of class_path or template_path can be defined")
        return values

    @root_validator(pre=True)
    def _validate_class_or_template(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Ensure that either class_path or template_path is defined."""
        # if the stack is disabled or locked, it is ok that these are missing
        required = values.get("enabled", True) and not values.get("locked", False)
        if (
            not values.get("class_path")
            and not values.get("template_path")
            and required
        ):
            raise ValueError("either class_path or template_path must be defined")
        return values


class CfnginConfigDefinitionModel(ConfigProperty):
    """Model for a CFNgin config definition."""

    cfngin_bucket: Optional[str] = Field(
        default=None,
        title="CFNgin Bucket",
        description="Name of an AWS S3 bucket to use for caching CloudFormation templates. "
        "Set as an empty string to disable caching.",
    )
    cfngin_bucket_region: Optional[str] = Field(
        default=None,
        title="CFNgin Bucket Region",
        description="AWS Region where the CFNgin Bucket is located. "
        "If not provided, the current region is used.",
    )
    cfngin_cache_dir: Optional[Path] = Field(
        default=None,
        title="CFNgin Cache Directory",
        description="Path to a local directory that CFNgin will use for local caching.",
    )
    log_formats: Dict[str, str] = Field(  # TODO create model
        default={}, description="Customize log message formatting by log level."
    )
    lookups: Dict[str, str] = Field(
        default={},
        description="Mapping of custom lookup names to a python importable path "
        "for the class that will be used to resolve the lookups.",
    )
    mappings: Dict[str, Dict[str, Dict[str, Any]]] = Field(
        default={}, description="Mappings that will be appended to all stack templates."
    )
    namespace: str = Field(
        ...,
        description="The namespace used to prefix stack names to create separation "
        "within an AWS account.",
    )
    namespace_delimiter: str = Field(
        default="-",
        description="Character used to separate the namespace and stack name "
        "when the namespace is prepended.",
    )
    package_sources: CfnginPackageSourcesDefinitionModel = Field(
        default=CfnginPackageSourcesDefinitionModel(),
        description=CfnginPackageSourcesDefinitionModel.Config.schema_extra[
            "description"
        ],
    )
    persistent_graph_key: Optional[str] = Field(
        default=None,
        description="Key for an AWS S3 object used to track a graph of stacks "
        "between executions.",
    )
    post_deploy: Union[
        List[CfnginHookDefinitionModel],  # final type after parsing
        Dict[str, CfnginHookDefinitionModel],  # recommended when writing config
    ] = Field(default=[], title="Post Deploy Hooks")
    post_destroy: Union[
        List[CfnginHookDefinitionModel],  # final type after parsing
        Dict[str, CfnginHookDefinitionModel],  # recommended when writing config
    ] = Field(default=[], title="Pre Destroy Hooks")
    pre_deploy: Union[
        List[CfnginHookDefinitionModel],  # final type after parsing
        Dict[str, CfnginHookDefinitionModel],  # recommended when writing config
    ] = Field(default=[], title="Pre Deploy Hooks")
    pre_destroy: Union[
        List[CfnginHookDefinitionModel],  # final type after parsing
        Dict[str, CfnginHookDefinitionModel],  # recommended when writing config
    ] = Field(default=[], title="Pre Destroy Hooks")
    service_role: Optional[str] = Field(
        default=None,
        title="Service Role ARN",
        description="Specify an IAM Role for CloudFormation to use.",
    )
    stacks: Union[
        List[CfnginStackDefinitionModel],  # final type after parsing
        Dict[str, CfnginStackDefinitionModel],  # recommended when writing config
    ] = Field(
        default=[],
        description="Define CloudFormation stacks using a Blueprint or Template.",
    )
    sys_path: Optional[Path] = Field(
        default=None,
        title="sys.path",
        description="Path to append to $PATH. This is also the root of relative paths.",
    )
    tags: Optional[Dict[str, str]] = Field(
        default=None,  # None is significant here
        description="Tags to try to apply to all resources created from this configuration file.",
    )
    template_indent: int = Field(
        default=4,
        description="Number of spaces per indentation level to use when "
        "rendering/outputting CloudFormation templates.",
    )

    class Config(ConfigProperty.Config):
        """Model configuration."""

        schema_extra = {"description": "Configuration file for Runway's CFNgin."}
        title = "CFNgin Config File"

    _resolve_path_fields = cast(
        "classmethod[Callable[..., Any]]",
        validator("cfngin_cache_dir", "sys_path", allow_reuse=True)(
            utils.resolve_path_field
        ),
    )

    @validator("post_deploy", "post_destroy", "pre_deploy", "pre_destroy", pre=True)
    def _convert_hook_definitions(
        cls, v: Union[Dict[str, Any], List[Dict[str, Any]]]
    ) -> List[Dict[str, Any]]:
        """Convert hooks defined as a dict to a list."""
        if isinstance(v, list):
            return v
        return list(v.values())

    @validator("stacks", pre=True)
    def _convert_stack_definitions(
        cls, v: Union[Dict[str, Any], List[Dict[str, Any]]]
    ) -> List[Dict[str, Any]]:
        """Convert stacks defined as a dict to a list."""
        if isinstance(v, list):
            return v
        result: List[Dict[str, Any]] = []
        for name, stack in copy.deepcopy(v).items():
            stack["name"] = name
            result.append(stack)
        return result

    @validator("stacks")
    def _validate_unique_stack_names(
        cls, stacks: List[CfnginStackDefinitionModel]
    ) -> List[CfnginStackDefinitionModel]:
        """Validate that each stack has a unique name."""
        stack_names = [stack.name for stack in stacks]
        if len(set(stack_names)) != len(stack_names):
            for i, name in enumerate(stack_names):
                if stack_names.count(name) != 1:
                    raise ValueError(f"Duplicate stack {name} found at index {i}")
        return stacks

    @classmethod
    def parse_file(
        cls: Type[Model],
        path: Union[str, Path],
        *,
        content_type: Optional[str] = None,
        encoding: str = "utf8",
        proto: Optional[Protocol] = None,
        allow_pickle: bool = False,
    ) -> Model:
        """Parse a file."""
        return cast(
            "Model",
            cls.parse_raw(
                Path(path).read_text(
                    encoding=locale.getpreferredencoding(do_setlocale=False)
                ),
                content_type=content_type,  # type: ignore
                encoding=encoding,
                proto=proto,  # type: ignore
                allow_pickle=allow_pickle,
            ),
        )

    @classmethod
    def parse_raw(
        cls: Type[Model],
        b: Union[bytes, str],
        *,
        content_type: Optional[str] = None,  # pylint: disable=unused-argument
        encoding: str = "utf8",  # pylint: disable=unused-argument
        proto: Optional[Protocol] = None,  # pylint: disable=unused-argument
        allow_pickle: bool = False,  # pylint: disable=unused-argument
    ) -> Model:
        """Parse raw data."""
        return cast("Model", cls.parse_obj(yaml.safe_load(b)))



"""Exact time stamps sync strategy.

.. Derived from software distributed by Amazon.com, Inc - http://aws.amazon.com/apache2.0/
   https://github.com/aws/aws-cli/blob/83b43782dd/awscli/customizations/s3/syncstrategy/exacttimestamps.py

"""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, ClassVar, Optional

from typing_extensions import Literal

from .base import SizeAndLastModifiedSync

if TYPE_CHECKING:
    from ..file_generator import FileStats

LOGGER = logging.getLogger(__name__.replace("._", "."))


class ExactTimestampsSync(SizeAndLastModifiedSync):
    """Exact time stamp."""

    NAME: ClassVar[Literal["exact_timestamps"]] = "exact_timestamps"

    def compare_time(
        self, src_file: Optional[FileStats], dest_file: Optional[FileStats]
    ) -> bool:
        """Compare modified time of two FileStats objects.

        Returns:
            True if the file does not need updating based on time of last
            modification and type of operation. False if the file does need
            updating based on the time of last modification and type of operation.

        """
        if not (src_file and dest_file):
            raise ValueError("src_file and dest_file must not be None")
        delta = dest_file.last_update - src_file.last_update
        if src_file.operation_name == "download":
            return delta.total_seconds() == 0
        return super().compare_time(src_file, dest_file)
